# -*- coding: utf-8 -*-
"""
Created on Wed Jul 22 14:04:30 2020

@author: Koerber
"""

from dl.method.mode import LearningMode, dlMode
from dl.method.pixelbasedprediction import PixelBasedPrediction
import utils.shapes.Contour as Contour
import dl.utils.dl_utils as dl_utils
import numpy as np
import cv2
from dl.loss.segmentation_losses import SegmentationLosses
from dl.metric.segmentation_metrics import SegmentationMetrics
from tensorflow.keras.utils import to_categorical


class Segmentation(PixelBasedPrediction, LearningMode):
    def __init__(self, parent):
        PixelBasedPrediction.__init__(self,parent)  
        LearningMode.__init__(self, parent)
        self.type = dlMode.Segmentation
        self.metric = SegmentationMetrics(parent)
        self.loss = SegmentationLosses(parent)
        self.architecture = 'Unet'
        self.backbone = 'resnet50'

    def LoadLabel(self, filename, height, width, rmBackground = False):
        label = np.zeros((height, width, 1), np.uint8)
        contours = Contour.loadContours(filename)
        label = Contour.drawContoursToLabel(label, contours)
        if rmBackground:
            return Contour.removeBackground(label)
        return label
    
    def addWeightMap(self, label):
        if self.parent.seg_useWeightedDistanceMap:
            weights = dl_utils.createWeightedBorderMapFromLabel(label[np.newaxis,...], w0 = self.parent.seg_border_weight, sigma = self.parent.seg_border_dist)
            label = np.concatenate((label, weights[0,...]), axis = 2)
        return label
    
    def prepreprocessLabel(self, label):
        return self.addWeightMap(label)
        
    def preprocessLabel(self, mask):
        if self.parent.NumClasses > 2:
            if self.parent.seg_useWeightedDistanceMap:                 
                label = mask[...,0]
                weights = mask[...,1]
                label = to_categorical(label, num_classes=self.parent.NumClasses)
                mask = np.concatenate((label, weights[...,np.newaxis]), axis = 3) 
            else:
                mask = to_categorical(mask, num_classes=self.parent.NumClasses)
        else:
            mask = mask.astype('float')
        return mask

    def getModel(self, nclasses, channels):
        if nclasses > 2:
            out_activation = 'softmax'
        else:
            out_activation =  'sigmoid'
        return PixelBasedPrediction.getPixelModel(self, nclasses, channels, out_activation)

    def extractShapesFromPrediction(self, prediction, innercontours, offset=(0,0)):
        return Contour.extractContoursFromLabel(prediction, innercontours, offset)
    
    def saveShapes(self, contours, path):
        Contour.saveContours(contours, path)
        
    def convert2Image(self,prediction):
        prediction = np.squeeze(prediction)
        
        if self.parent.seg_wt_separation:
            return dl_utils.separatePredictions(prediction, self.parent.seg_wt_mindist, self.parent.seg_wt_thresh)
        
        if self.parent.NumClasses > 2:
            prediction = np.squeeze(np.argmax(prediction, axis = 2))
        else:
            prediction[prediction>0.5] = 1
            prediction[prediction<=0.5] = 0   
        return prediction

    def LabelDistance(self,l1, l2):
        return Contour.matchcontours(l1,l2)

    def LoadShapes(self,filename):
        contours = Contour.loadContours(filename)
        return contours
    
    def resizeLabel(self, label, shape):
        return cv2.resize(label, shape, interpolation = cv2.INTER_NEAREST)

    def countLabelWeight(self, label):
        classValues = np.zeros((self.parent.NumClasses_real), dtype = np.float)
        for i in range(self.parent.NumClasses_real):
            classValues[i] = np.count_nonzero((label == i))
        return classValues


# -*- coding: utf-8 -*-
"""
Created on Thu Jul  9 11:30:13 2020

@author: Koerber
"""

from utils.Image import supportedImageFormats
import glob
import os
import time
import re
from shutil import copyfile


class FilesAndFolders():
    def __init__(self, parent):
        self.parent = parent
        
        self.trainImagespath = None
        self.validationpath = None
        self.testImagespath = None
        self.subFolder = None

        self.train_test_dir = True # True = TrainFolder; False = TestFolder
        
        self.files = None
        self.currentImage = -1
        self.numofImages = 0
        self.currentFrame = 0

    @property
    def currentimagesrootpath(self): # naming unclear
        path = self.trainImagespath if self.train_test_dir else self.testImagespath
        return path 
        
    @property
    def currentimagespath(self): # naming unclear
        path = self.currentimagesrootpath
        if self.subFolder:
            path = path + self.subFolder
        return path 
        
    @property
    def currentlabelspath(self): # naming unclear
        path = self.currentimagespath
        if not path:
            return
        path = path + os.path.sep + self.parent.dl.LabelFolderName
        if self.parent.hasStack():
            path = self.extendLabelPathByFolder(path, self.CurrentImageName()) 
        return path
    
    @property
    # not used anymore 
    def trainImageLabelspath(self):
        if self.trainImagespath:
            return self.trainImagespath + os.path.sep + self.parent.dl.LabelFolderName
       
    @property
    def testImageLabelspath(self):
        if self.testImagespath:
            return self.testImagespath + os.path.sep + self.parent.dl.LabelFolderName
        
    # @property
    def CurrentImageName(self, withext = False): 
        if self.CurrentImagePath() is None:
            return None
        return self.getFilenameFromPath(self.CurrentImagePath(), withext)

    # @property
    def CurrentLabelName(self):
        if self.parent.hasStack():
            return self.extendNameByFrameNumber(self.CurrentImageName(), self.currentFrame)
        else:
            return self.CurrentImageName()
    # @property
    def CurrentImagePath(self):
        if self.files is None:
            return None
        return self.files[self.currentImage]
        
    # @property
    def CurrentLabelPath(self):
        if self.currentlabelspath is None or self.CurrentImageName() is None:
            return None
        path = os.path.join(self.currentlabelspath, self.CurrentImageName())
        if self.parent.hasStack() and self.parent.separateStackLabels:
            return self.extendNameByFrameNumber(path, self.currentFrame) + ".npz"
        else:
            return path + ".npz"

    @property
    def TestImages(self):
        return self.getImagesFromFolder(self.testImagespath)

    def nextImage(self):
        self.currentImage += 1
        if (self.currentImage >= self.numofImages):
            self.currentImage = 0
    
    def previousImage(self):
        self.currentImage -= 1
        if (self.currentImage < 0):
            self.currentImage = self.numofImages - 1
            
    def setTrainFolder(self, folder):
        self.trainImagespath = folder
        self.subFolder = None
        self.train_test_dir = True
        
    def setTestFolder(self, folder):
        self.testImagespath = folder
        self.subFolder = None
        self.train_test_dir = False
     
    def switchToTrainFolder(self):
        self.train_test_dir = True
        
    def switchToTestFolder(self):
        self.train_test_dir = False          
        
    def ensureLabelFolder(self):
        self.ensureFolder(self.currentlabelspath)
        
    def getFiles(self):
        if self.currentimagespath is None:
            return
        self.files = self.getImagesFromFolder(self.currentimagespath)
        filenames = [self.getFilenameFromPath(x) for x in self.files]
        
        if len(filenames) != len(set(filenames)):
            self.parent.PopupWarning('Image names contain duplicates (without file extension). Please rename images to avoid duplicates')
            self.files = []

        if not self.files:
            self.currentImage = -1
            self.files = None
            self.numofImages = 0
        else:
            self.sortNatural(self.files)
            self.currentImage = 0
            self.numofImages = len(self.files)

    def moveToSubFolder(self, path):
        curr_path = os.path.abspath(self.currentimagesrootpath)
        path = os.path.abspath(path)
        if curr_path not in path:
            return False
        subfolder = path.replace(curr_path,'')

        if subfolder != self.subFolder:
            self.subFolder = subfolder
            self.getFiles()
            return True
        return False

    def setImagebyName(self,filename):
        if not self.files:
            return
        filenames = [self.getFilenameFromPath(x,True) for x in self.files]
        try:
            self.currentImage = filenames.index(filename)
            return True
        except:
            return False
        
    ## helper functions
    def sortNatural(self, alist):
        def atoi(text):
            return int(text) if text.isdigit() else text
        def natural_keys(text):
            # http://nedbatchelder.com/blog/200712/human_sorting.html
            return [ atoi(c) for c in re.split(r'(\d+)', text) ]
        if alist:
            return alist.sort(key=natural_keys)

    def getImagesFromFolder(self, folder):
        images = glob.glob(os.path.join(folder,'*.*'))
        images = [x for x in images if (x.lower().endswith(tuple(supportedImageFormats())))]
        self.sortNatural(images)
        return images

    def getFilenameFromPath(self, path, withfileextension=False):
        file = os.path.splitext(os.path.basename(path))
        if withfileextension:
            return ''.join(file)
        else:
            return file[0]
        
    def extendLabelPathByFolder(self, path, folder):
        return os.path.join(path, folder) 

    def extendNameByFrameNumber(self,name, frame):
        return name + '_' + "{0:0=3d}".format(frame)

    def getFrameNumber(self, labelpath):
        if self.isStackLabel(labelpath):
            return int(labelpath[1])
        labelpath = self.convertIfStackPath(labelpath)
        return int(labelpath[-7:-4])

    def convertIfStackPath(self, path):
        if self.isStackLabel(path):
            return path[0]
        return path

    def isStackLabel(self,path):
        if isinstance(path,tuple):
            return True
        else:
            return False

    def convertLabelPath2FileName(self, labelpath):
        if self.isStackLabel(labelpath):
            name = os.path.splitext(os.path.basename(labelpath[0]))[0][:-3] + str(int(labelpath[1])+1)
        else:
            name = os.path.splitext(os.path.basename(labelpath))[0]
        return name
         
    def ensureFolder(self, foldername):
        if not os.path.isdir(foldername):
            os.makedirs(foldername)
            # os.mkdir(foldername)

    def removeFrameNumFromLabelName(self, stacklabel):
        return stacklabel[:-4]

    def ImagePath2LabelPath(self, imagepath, checkexistence = False, stack = False):
        if imagepath is None:
            return None
        filename = self.getFilenameFromPath(imagepath)
        path = os.path.join(os.path.dirname(imagepath), self.parent.dl.LabelFolderName, filename)
        if not checkexistence:
            if stack:
                return glob.glob(os.path.join(path,'*.*'))
            else:
                return path +  ".npz"
        else:
            if stack:
                return glob.glob(os.path.join(path,'*.*'))
            else:
                path += ".npz"
                return path if os.path.exists(path) else None

    def copyStackLabels(self, label, frames):
        print(label)
        print(frames)
        if not self.parent.hasStack():
            return
        label = self.convertIfStackPath(label)
        folder = os.path.dirname(os.path.abspath(label))
        name = self.removeFrameNumFromLabelName(self.getFilenameFromPath(label))
        _,ext = os.path.splitext(label)
        for f in range(frames):
            if f ==  self.getFrameNumber(label):
                continue
            else:
                filename = self.extendNameByFrameNumber(name,f) + ext
                copyfile(label, os.path.join(folder,filename))


"""ResNet, ResNetV2, and ResNeXt models for Keras.

# Reference papers

- [Deep Residual Learning for Image Recognition]
  (https://arxiv.org/abs/1512.03385) (CVPR 2016 Best Paper Award)
- [Identity Mappings in Deep Residual Networks]
  (https://arxiv.org/abs/1603.05027) (ECCV 2016)
- [Aggregated Residual Transformations for Deep Neural Networks]
  (https://arxiv.org/abs/1611.05431) (CVPR 2017)

# Reference implementations

- [TensorNets]
  (https://github.com/taehoonlee/tensornets/blob/master/tensornets/resnets.py)
- [Caffe ResNet]
  (https://github.com/KaimingHe/deep-residual-networks/tree/master/prototxt)
- [Torch ResNetV2]
  (https://github.com/facebook/fb.resnet.torch/blob/master/models/preresnet.lua)
- [Torch ResNeXt]
  (https://github.com/facebookresearch/ResNeXt/blob/master/models/resnext.lua)

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from . import get_submodules_from_kwargs
from .imagenet_utils import _obtain_input_shape
from dl.utils.dl_downloads import get_ResNet

backend = None
layers = None
models = None
keras_utils = None




def block1(x, filters, kernel_size=3, stride=1,
           conv_shortcut=True, name=None):
    """A residual block.

    # Arguments
        x: input tensor.
        filters: integer, filters of the bottleneck layer.
        kernel_size: default 3, kernel size of the bottleneck layer.
        stride: default 1, stride of the first layer.
        conv_shortcut: default True, use convolution shortcut if True,
            otherwise identity shortcut.
        name: string, block label.

    # Returns
        Output tensor for the residual block.
    """
    bn_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    if conv_shortcut is True:
        shortcut = layers.Conv2D(4 * filters, 1, strides=stride,
                                 name=name + '_0_conv')(x)
        shortcut = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                             name=name + '_0_bn')(shortcut)
    else:
        shortcut = x

    x = layers.Conv2D(filters, 1, strides=stride, name=name + '_1_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                  name=name + '_1_bn')(x)
    x = layers.Activation('relu', name=name + '_1_relu')(x)

    x = layers.Conv2D(filters, kernel_size, padding='SAME',
                      name=name + '_2_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                  name=name + '_2_bn')(x)
    x = layers.Activation('relu', name=name + '_2_relu')(x)

    x = layers.Conv2D(4 * filters, 1, name=name + '_3_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                  name=name + '_3_bn')(x)

    x = layers.Add(name=name + '_add')([shortcut, x])
    x = layers.Activation('relu', name=name + '_out')(x)
    return x


def stack1(x, filters, blocks, stride1=2, name=None):
    """A set of stacked residual blocks.

    # Arguments
        x: input tensor.
        filters: integer, filters of the bottleneck layer in a block.
        blocks: integer, blocks in the stacked blocks.
        stride1: default 2, stride of the first layer in the first block.
        name: string, stack label.

    # Returns
        Output tensor for the stacked blocks.
    """
    x = block1(x, filters, stride=stride1, name=name + '_block1')
    for i in range(2, blocks + 1):
        x = block1(x, filters, conv_shortcut=False, name=name + '_block' + str(i))
    return x


def block2(x, filters, kernel_size=3, stride=1,
           conv_shortcut=False, name=None):
    """A residual block.

    # Arguments
        x: input tensor.
        filters: integer, filters of the bottleneck layer.
        kernel_size: default 3, kernel size of the bottleneck layer.
        stride: default 1, stride of the first layer.
        conv_shortcut: default False, use convolution shortcut if True,
            otherwise identity shortcut.
        name: string, block label.

    # Returns
        Output tensor for the residual block.
    """
    bn_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    preact = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                       name=name + '_preact_bn')(x)
    preact = layers.Activation('relu', name=name + '_preact_relu')(preact)

    if conv_shortcut is True:
        shortcut = layers.Conv2D(4 * filters, 1, strides=stride,
                                 name=name + '_0_conv')(preact)
    else:
        shortcut = layers.MaxPooling2D(1, strides=stride)(x) if stride > 1 else x

    x = layers.Conv2D(filters, 1, strides=1, use_bias=False,
                      name=name + '_1_conv')(preact)
    x = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                  name=name + '_1_bn')(x)
    x = layers.Activation('relu', name=name + '_1_relu')(x)

    x = layers.ZeroPadding2D(padding=((1, 1), (1, 1)), name=name + '_2_pad')(x)
    x = layers.Conv2D(filters, kernel_size, strides=stride,
                      use_bias=False, name=name + '_2_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                  name=name + '_2_bn')(x)
    x = layers.Activation('relu', name=name + '_2_relu')(x)

    x = layers.Conv2D(4 * filters, 1, name=name + '_3_conv')(x)
    x = layers.Add(name=name + '_out')([shortcut, x])
    return x


def stack2(x, filters, blocks, stride1=2, name=None):
    """A set of stacked residual blocks.

    # Arguments
        x: input tensor.
        filters: integer, filters of the bottleneck layer in a block.
        blocks: integer, blocks in the stacked blocks.
        stride1: default 2, stride of the first layer in the first block.
        name: string, stack label.

    # Returns
        Output tensor for the stacked blocks.
    """
    x = block2(x, filters, conv_shortcut=True, name=name + '_block1')
    for i in range(2, blocks):
        x = block2(x, filters, name=name + '_block' + str(i))
    x = block2(x, filters, stride=stride1, name=name + '_block' + str(blocks))
    return x


def block3(x, filters, kernel_size=3, stride=1, groups=32,
           conv_shortcut=True, name=None):
    """A residual block.

    # Arguments
        x: input tensor.
        filters: integer, filters of the bottleneck layer.
        kernel_size: default 3, kernel size of the bottleneck layer.
        stride: default 1, stride of the first layer.
        groups: default 32, group size for grouped convolution.
        conv_shortcut: default True, use convolution shortcut if True,
            otherwise identity shortcut.
        name: string, block label.

    # Returns
        Output tensor for the residual block.
    """
    bn_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    if conv_shortcut is True:
        shortcut = layers.Conv2D((64 // groups) * filters, 1, strides=stride,
                                 use_bias=False, name=name + '_0_conv')(x)
        shortcut = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                             name=name + '_0_bn')(shortcut)
    else:
        shortcut = x

    x = layers.Conv2D(filters, 1, use_bias=False, name=name + '_1_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                  name=name + '_1_bn')(x)
    x = layers.Activation('relu', name=name + '_1_relu')(x)

    c = filters // groups
    x = layers.ZeroPadding2D(padding=((1, 1), (1, 1)), name=name + '_2_pad')(x)
    x = layers.DepthwiseConv2D(kernel_size, strides=stride, depth_multiplier=c,
                               use_bias=False, name=name + '_2_conv')(x)
    x_shape = backend.int_shape(x)[1:-1]
    x = layers.Reshape(x_shape + (groups, c, c))(x)
    output_shape = x_shape + (groups, c) if backend.backend() == 'theano' else None
    x = layers.Lambda(lambda x: sum([x[:, :, :, :, i] for i in range(c)]),
                      output_shape=output_shape, name=name + '_2_reduce')(x)
    x = layers.Reshape(x_shape + (filters,))(x)
    x = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                  name=name + '_2_bn')(x)
    x = layers.Activation('relu', name=name + '_2_relu')(x)

    x = layers.Conv2D((64 // groups) * filters, 1,
                      use_bias=False, name=name + '_3_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                  name=name + '_3_bn')(x)

    x = layers.Add(name=name + '_add')([shortcut, x])
    x = layers.Activation('relu', name=name + '_out')(x)
    return x


def stack3(x, filters, blocks, stride1=2, groups=32, name=None):
    """A set of stacked residual blocks.

    # Arguments
        x: input tensor.
        filters: integer, filters of the bottleneck layer in a block.
        blocks: integer, blocks in the stacked blocks.
        stride1: default 2, stride of the first layer in the first block.
        groups: default 32, group size for grouped convolution.
        name: string, stack label.

    # Returns
        Output tensor for the stacked blocks.
    """
    x = block3(x, filters, stride=stride1, groups=groups, name=name + '_block1')
    for i in range(2, blocks + 1):
        x = block3(x, filters, groups=groups, conv_shortcut=False,
                   name=name + '_block' + str(i))
    return x


def ResNet(stack_fn,
           preact,
           use_bias,
           model_name='resnet',
           include_top=True,
           weights='imagenet',
           input_tensor=None,
           input_shape=None,
           pooling=None,
           classes=1000,
           **kwargs):
    """Instantiates the ResNet, ResNetV2, and ResNeXt architecture.

    Optionally loads weights pre-trained on ImageNet.
    Note that the data format convention used by the model is
    the one specified in your Keras config at `~/.keras/keras.json`.

    # Arguments
        stack_fn: a function that returns output tensor for the
            stacked residual blocks.
        preact: whether to use pre-activation or not
            (True for ResNetV2, False for ResNet and ResNeXt).
        use_bias: whether to use biases for convolutional layers or not
            (True for ResNet and ResNetV2, False for ResNeXt).
        model_name: string, model name.
        include_top: whether to include the fully-connected
            layer at the top of the network.
        weights: one of `None` (random initialization),
              'imagenet' (pre-training on ImageNet),
              or the path to the weights file to be loaded.
        input_tensor: optional Keras tensor
            (i.e. output of `layers.Input()`)
            to use as image input for the model.
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False (otherwise the input shape
            has to be `(224, 224, 3)` (with `channels_last` data format)
            or `(3, 224, 224)` (with `channels_first` data format).
            It should have exactly 3 inputs channels.
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be
                the 4D tensor output of the
                last convolutional layer.
            - `avg` means that global average pooling
                will be applied to the output of the
                last convolutional layer, and thus
                the output of the model will be a 2D tensor.
            - `max` means that global max pooling will
                be applied.
        classes: optional number of classes to classify images
            into, only to be specified if `include_top` is True, and
            if no `weights` argument is specified.

    # Returns
        A Keras model instance.

    # Raises
        ValueError: in case of invalid argument for `weights`,
            or invalid input shape.
    """
    global backend, layers, models, keras_utils
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)

    if not (weights in {'imagenet', None} or os.path.exists(weights)):
        raise ValueError('The `weights` argument should be either '
                         '`None` (random initialization), `imagenet` '
                         '(pre-training on ImageNet), '
                         'or the path to the weights file to be loaded.')

    if weights == 'imagenet' and include_top and classes != 1000:
        raise ValueError('If using `weights` as `"imagenet"` with `include_top`'
                         ' as true, `classes` should be 1000')

    # Determine proper input shape
    input_shape = _obtain_input_shape(input_shape,
                                      default_size=224,
                                      min_size=32,
                                      data_format=backend.image_data_format(),
                                      require_flatten=include_top,
                                      weights=weights)

    if input_tensor is None:
        img_input = layers.Input(shape=input_shape)
    else:
        if not backend.is_keras_tensor(input_tensor):
            img_input = layers.Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    bn_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    x = layers.ZeroPadding2D(padding=((3, 3), (3, 3)), name='conv1_pad')(img_input)
    x = layers.Conv2D(64, 7, strides=2, use_bias=use_bias, name='conv1_conv')(x)

    if preact is False:
        x = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                      name='conv1_bn')(x)
        x = layers.Activation('relu', name='conv1_relu')(x)

    x = layers.ZeroPadding2D(padding=((1, 1), (1, 1)), name='pool1_pad')(x)
    x = layers.MaxPooling2D(3, strides=2, name='pool1_pool')(x)

    x = stack_fn(x)

    if preact is True:
        x = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5,
                                      name='post_bn')(x)
        x = layers.Activation('relu', name='post_relu')(x)

    if include_top:
        x = layers.GlobalAveragePooling2D(name='avg_pool')(x)
        x = layers.Dense(classes, activation='softmax', name='probs')(x)
    else:
        if pooling == 'avg':
            x = layers.GlobalAveragePooling2D(name='avg_pool')(x)
        elif pooling == 'max':
            x = layers.GlobalMaxPooling2D(name='max_pool')(x)

    # Ensure that the model takes into account
    # any potential predecessors of `input_tensor`.
    if input_tensor is not None:
        inputs = keras_utils.get_source_inputs(input_tensor)
    else:
        inputs = img_input

    # Create model.
    model = models.Model(inputs, x, name=model_name)

    # Load weights.
    if (weights == 'imagenet') and (model_name in WEIGHTS_HASHES):
        weights_path = get_ResNet(include_top, model_name)
        model.load_weights(weights_path)
    elif weights is not None:
        model.load_weights(weights)

    return model


def ResNet50(include_top=True,
             weights='imagenet',
             input_tensor=None,
             input_shape=None,
             pooling=None,
             classes=1000,
             **kwargs):
    def stack_fn(x):
        x = stack1(x, 64, 3, stride1=1, name='conv2')
        x = stack1(x, 128, 4, name='conv3')
        x = stack1(x, 256, 6, name='conv4')
        x = stack1(x, 512, 3, name='conv5')
        return x
    return ResNet(stack_fn, False, True, 'resnet50',
                  include_top, weights,
                  input_tensor, input_shape,
                  pooling, classes,
                  **kwargs)


def ResNet101(include_top=True,
              weights='imagenet',
              input_tensor=None,
              input_shape=None,
              pooling=None,
              classes=1000,
              **kwargs):
    def stack_fn(x):
        x = stack1(x, 64, 3, stride1=1, name='conv2')
        x = stack1(x, 128, 4, name='conv3')
        x = stack1(x, 256, 23, name='conv4')
        x = stack1(x, 512, 3, name='conv5')
        return x
    return ResNet(stack_fn, False, True, 'resnet101',
                  include_top, weights,
                  input_tensor, input_shape,
                  pooling, classes,
                  **kwargs)


def ResNet152(include_top=True,
              weights='imagenet',
              input_tensor=None,
              input_shape=None,
              pooling=None,
              classes=1000,
              **kwargs):
    def stack_fn(x):
        x = stack1(x, 64, 3, stride1=1, name='conv2')
        x = stack1(x, 128, 8, name='conv3')
        x = stack1(x, 256, 36, name='conv4')
        x = stack1(x, 512, 3, name='conv5')
        return x
    return ResNet(stack_fn, False, True, 'resnet152',
                  include_top, weights,
                  input_tensor, input_shape,
                  pooling, classes,
                  **kwargs)


def ResNet50V2(include_top=True,
               weights='imagenet',
               input_tensor=None,
               input_shape=None,
               pooling=None,
               classes=1000,
               **kwargs):
    def stack_fn(x):
        x = stack2(x, 64, 3, name='conv2')
        x = stack2(x, 128, 4, name='conv3')
        x = stack2(x, 256, 6, name='conv4')
        x = stack2(x, 512, 3, stride1=1, name='conv5')
        return x
    return ResNet(stack_fn, True, True, 'resnet50v2',
                  include_top, weights,
                  input_tensor, input_shape,
                  pooling, classes,
                  **kwargs)


def ResNet101V2(include_top=True,
                weights='imagenet',
                input_tensor=None,
                input_shape=None,
                pooling=None,
                classes=1000,
                **kwargs):
    def stack_fn(x):
        x = stack2(x, 64, 3, name='conv2')
        x = stack2(x, 128, 4, name='conv3')
        x = stack2(x, 256, 23, name='conv4')
        x = stack2(x, 512, 3, stride1=1, name='conv5')
        return x
    return ResNet(stack_fn, True, True, 'resnet101v2',
                  include_top, weights,
                  input_tensor, input_shape,
                  pooling, classes,
                  **kwargs)


def ResNet152V2(include_top=True,
                weights='imagenet',
                input_tensor=None,
                input_shape=None,
                pooling=None,
                classes=1000,
                **kwargs):
    def stack_fn(x):
        x = stack2(x, 64, 3, name='conv2')
        x = stack2(x, 128, 8, name='conv3')
        x = stack2(x, 256, 36, name='conv4')
        x = stack2(x, 512, 3, stride1=1, name='conv5')
        return x
    return ResNet(stack_fn, True, True, 'resnet152v2',
                  include_top, weights,
                  input_tensor, input_shape,
                  pooling, classes,
                  **kwargs)


def ResNeXt50(include_top=True,
              weights='imagenet',
              input_tensor=None,
              input_shape=None,
              pooling=None,
              classes=1000,
              **kwargs):
    def stack_fn(x):
        x = stack3(x, 128, 3, stride1=1, name='conv2')
        x = stack3(x, 256, 4, name='conv3')
        x = stack3(x, 512, 6, name='conv4')
        x = stack3(x, 1024, 3, name='conv5')
        return x
    return ResNet(stack_fn, False, False, 'resnext50',
                  include_top, weights,
                  input_tensor, input_shape,
                  pooling, classes,
                  **kwargs)


def ResNeXt101(include_top=True,
               weights='imagenet',
               input_tensor=None,
               input_shape=None,
               pooling=None,
               classes=1000,
               **kwargs):
    def stack_fn(x):
        x = stack3(x, 128, 3, stride1=1, name='conv2')
        x = stack3(x, 256, 4, name='conv3')
        x = stack3(x, 512, 23, name='conv4')
        x = stack3(x, 1024, 3, name='conv5')
        return x
    return ResNet(stack_fn, False, False, 'resnext101',
                  include_top, weights,
                  input_tensor, input_shape,
                  pooling, classes,
                  **kwargs)


setattr(ResNet50, '__doc__', ResNet.__doc__)
setattr(ResNet101, '__doc__', ResNet.__doc__)
setattr(ResNet152, '__doc__', ResNet.__doc__)
setattr(ResNet50V2, '__doc__', ResNet.__doc__)
setattr(ResNet101V2, '__doc__', ResNet.__doc__)
setattr(ResNet152V2, '__doc__', ResNet.__doc__)
setattr(ResNeXt50, '__doc__', ResNet.__doc__)
setattr(ResNeXt101, '__doc__', ResNet.__doc__)



from . import get_submodules_from_kwargs

from dl.utils.dl_downloads import WEIGHTS_COLLECTION, getTargetFile

__all__ = ['load_model_weights']


def _find_weights(model_name, dataset, include_top):
    w = list(filter(lambda x: x['model'] == model_name, WEIGHTS_COLLECTION))
    w = list(filter(lambda x: x['dataset'] == dataset, w))
    w = list(filter(lambda x: x['include_top'] == include_top, w))
    return w


def load_model_weights(model, model_name, dataset, classes, include_top, **kwargs):
    _, _, _, keras_utils = get_submodules_from_kwargs(kwargs)

    weights = _find_weights(model_name, dataset, include_top)

    if weights:
        weights = weights[0]

        if include_top and weights['classes'] != classes:
            raise ValueError('If using `weights` and `include_top`'
                             ' as true, `classes` should be {}'.format(weights['classes']))

        weights_path = getTargetFile(weights['name'], weights['url'], weights['md5'])

        model.load_weights(weights_path)

    else:
        raise ValueError('There is no weights for such configuration: ' +
                         'model = {}, dataset = {}, '.format(model.name, dataset) +
                         'classes = {}, include_top = {}.'.format(classes, include_top))




# -*- coding: utf-8 -*-

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
 
   
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from ui.style import styleForm, getHighlightColor, getBackgroundColor
from ui.ui_utils import DCDButton, saveFile
import math
import threading

class Window(object):
    def setupUi(self, Form):
        Form.setWindowTitle('Training Data') 
        styleForm(Form)
        self.centralWidget = QWidget(Form)
        layout = QVBoxLayout(self.centralWidget)
        self.centralWidget.setLayout(layout)

        self.figure = Figure(constrained_layout = True, dpi= 100)
        self.canvas = FigureCanvas(self.figure)
        self.figure.set_size_inches(6, 6)

        self.BStop = DCDButton(self.centralWidget,'Stop Training')
        self.BStop.setIcon(QIcon('icons/train.png'))
        self.BSave = DCDButton(self.centralWidget)
        self.BSave.setIcon(QIcon('icons/savemodel.png'))
        self.BSave.setToolTip('Save training data')
        
        
        l1 = QHBoxLayout()

        self.LBatch = QLabel(self.centralWidget)
        self.LEpoch = QLabel(self.centralWidget)

        l1.addWidget(self.LBatch, stretch=99, alignment= Qt.AlignLeft)
        l1.addWidget(self.LEpoch, stretch=99, alignment= Qt.AlignRight)
        l1.addWidget(self.BSave, stretch=0)


        layout.addLayout(l1)
        layout.addWidget(self.canvas)  
        
        l2 = QHBoxLayout()
        l2.addWidget(self.BStop)

        
        layout.addLayout(l2)

        self.centralWidget.setFixedSize(layout.sizeHint())
        Form.setFixedSize(layout.sizeHint())

class TrainPlotWindow(QMainWindow, Window):
    def __init__(self,parent):
        super(TrainPlotWindow, self).__init__(parent)
        self.setupUi(self)
        self.parent = parent
        self.dl = parent.dl
        self.lock = threading.Lock()
        
        self.train_loss_axis = None
        self.train_metric_axis = None
        
        self.val_loss_axis = None
        self.val_metric_axis = None
        
        self.BStop.clicked.connect(self.Stop)
        self.BSave.clicked.connect(self.saveTrainingRecord)
        
    def hide(self):
        self.dl.interrupt()
        super(TrainPlotWindow, self).hide()
        
    def closeEvent(self, event):       
        self.dl.interrupt()
        super(TrainPlotWindow, self).closeEvent(event)
        
    def initialize(self):
        bg = getBackgroundColor()
        self.figure.patch.set_facecolor((bg.red()/255, bg.green()/255, bg.blue()/255))
        color = 'white'
        
        plt.rcParams['text.color'] = color
        plt.rcParams['axes.labelcolor'] = color
        plt.rcParams['axes.edgecolor'] = color
        plt.rcParams['xtick.color'] = color
        plt.rcParams['ytick.color'] = color

        
        if self.train_loss_axis is None or self.train_metric_axis is None: 
            gs = self.figure.add_gridspec(2, 2)
            self.train_loss_axis = self.figure.add_subplot(gs[0, 0])
            self.train_metric_axis = self.figure.add_subplot(gs[0, 1])
            self.train_loss_axis.set_facecolor((0.0, 0.0, 0.0))
            self.train_metric_axis.set_facecolor((0.0, 0.0, 0.0))
            
            
        if self.val_loss_axis is None or self.val_metric_axis is None: 
            self.val_loss_axis = self.figure.add_subplot(gs[1, 0])
            self.val_metric_axis = self.figure.add_subplot(gs[1, 1])
            self.val_loss_axis.set_facecolor((0.0, 0.0, 0.0))
            self.val_metric_axis.set_facecolor((0.0, 0.0, 0.0))

        # self.canvas.draw()
        self.refresh()

        
    def updatePlot(self):
        # to do make drawing faster, ie using blitting or animations
        # but be careful, there seems to be some memory leaks in matplotlib blitting
        # self.canvas.draw() is the time consuming part
        
        if self.lock.locked():
            return

        with self.lock:

            loss = self.dl.record.loss
            metric =  self.dl.record.metric
            val_loss = self.dl.record.val_loss
            val_metric = self.dl.record.val_metric
            lpe = self.dl.record.loss_per_epoch
            mpe = self.dl.record.metric_per_epoch

            
            self.train_loss_axis.clear()
            self.train_metric_axis.clear()

            
            if len(loss) > 0:
                self.train_loss_axis.set_ylabel('loss')
                self.train_loss_axis.set_xlabel('iteration')
    
            if len(self.dl.Model.metrics_names) > 1:
                self.train_metric_axis.set_ylabel(self.dl.Model.metrics_names[1])
                self.train_metric_axis.set_xlabel('iteration')
                
            if self.dl.data.validationData():
                self.val_loss_axis.set_visible(True)
                self.val_metric_axis.set_visible(True)
                self.val_loss_axis.clear()
                self.val_metric_axis.clear()
                self.val_loss_axis.set_ylabel('validation loss')
                self.val_loss_axis.set_xlabel('epoch')
               
                if len(self.dl.Model.metrics_names) > 1:
                    self.val_metric_axis.set_ylabel('validation ' + self.dl.Model.metrics_names[1])
                    self.val_metric_axis.set_xlabel('epoch')
            else:
                self.val_loss_axis.set_visible(False)
                self.val_metric_axis.set_visible(False)
                
            col = getHighlightColor()
            color = (col.red()/255, col.green()/255, col.blue()/255)

            self.train_loss_axis.plot(loss, '-' if len(loss) > 1 else '.', color = color)
            self.train_metric_axis.plot(metric, '-' if len(metric) > 1 else '.', color = color)  
            self.val_loss_axis.plot(val_loss, '-' if len(val_loss) > 1 else '.',  color='red' )
            self.val_loss_axis.plot(lpe, '--' if len(lpe) > 1 else '.', color=color )
            self.val_metric_axis.plot(val_metric, '-' if len(val_metric) > 1 else '.', color = 'red')
            self.val_metric_axis.plot(mpe, '--' if len(mpe) > 1 else '.', color = color)

    
            self.LBatch.setText("Batch: %d of %d" % (min(self.dl.data.getNumberOfBatches(),self.dl.record.currentbatch +1), self.dl.data.getNumberOfBatches()))
            self.LEpoch.setText("Epoch: %d of %d" % (min(self.dl.epochs,self.dl.record.currentepoch +1), self.dl.epochs))
            
            self.canvas.draw()


    def saveTrainingRecord(self):
        filename = saveFile("Save Results","Comma Separated File (*.csv)", 'csv') 
        if not filename:
            return
        try:
            self.dl.record.save(filename)  

        except PermissionError: 
            self.parent.emitPopup('Cannot write file (already open?)')
            return

          
    def refresh(self):
        
        thread = threading.Thread(target=self.updatePlot)
        thread.daemon = True
        thread.start()
        
        # self.updatePlot()

    def Stop(self):
        self.dl.interrupt()

    def Resume(self):
        self.parent.resumeTraining()

    def toggleTrainStatus(self, training):
        self.BStop.disconnect()
        if training:
            self.BStop.setText('Stop Training')
            self.BStop.setIcon(QIcon('icons/delete.png'))
            self.BStop.clicked.connect(self.Stop)
        else:
            self.BStop.setText('Resume Training')
            self.BStop.setIcon(QIcon('icons/train.png'))
            self.BStop.clicked.connect(self.Resume)



"""ResNetV2 models for Keras.

# Reference paper

- [Aggregated Residual Transformations for Deep Neural Networks]
  (https://arxiv.org/abs/1611.05431) (CVPR 2017)

# Reference implementations

- [TensorNets]
  (https://github.com/taehoonlee/tensornets/blob/master/tensornets/resnets.py)
- [Torch ResNetV2]
  (https://github.com/facebook/fb.resnet.torch/blob/master/models/preresnet.lua)

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from . import imagenet_utils
from .imagenet_utils import decode_predictions
from .resnet_common import ResNet50V2
from .resnet_common import ResNet101V2
from .resnet_common import ResNet152V2


def preprocess_input(x, **kwargs):
    """Preprocesses a numpy array encoding a batch of images.

    # Arguments
        x: a 4D numpy array consists of RGB values within [0, 255].
        data_format: data format of the image tensor.

    # Returns
        Preprocessed array.
    """
    return imagenet_utils.preprocess_input(x, mode='tf', **kwargs)



from . import inject_keras_modules, init_keras_custom_objects
from . import model

from .preprocessing import center_crop_and_resize

EfficientNetB0 = inject_keras_modules(model.EfficientNetB0)
EfficientNetB1 = inject_keras_modules(model.EfficientNetB1)
EfficientNetB2 = inject_keras_modules(model.EfficientNetB2)
EfficientNetB3 = inject_keras_modules(model.EfficientNetB3)
EfficientNetB4 = inject_keras_modules(model.EfficientNetB4)
EfficientNetB5 = inject_keras_modules(model.EfficientNetB5)
EfficientNetB6 = inject_keras_modules(model.EfficientNetB6)
EfficientNetB7 = inject_keras_modules(model.EfficientNetB7)

preprocess_input = inject_keras_modules(model.preprocess_input)

init_keras_custom_objects()



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 16:44:23 2020

@author: Koerber
"""

from dl.loss.losses import Losses, dlLoss
from dl.loss.loss_functions import *

class ClassificationLosses(Losses):
    def __init__(self,parent):
        super(ClassificationLosses,self).__init__(parent)
        self.supportedlosses.append(dlLoss.focal_loss)
        self.supportedlosses.append(dlLoss.cross_entropy)
        self.supportedlosses.append(dlLoss.kl_divergence)
        self.loss = dlLoss.cross_entropy
        
    def getLoss(self):
        if self.loss == dlLoss.focal_loss:
            return self.focal_Loss_function(binary = self.parent.NumClasses <= 2, class_weights=self.parent.data.class_weights)
        if self.loss == dlLoss.cross_entropy:
            return self.categorical_cross_entropy_loss_function(binary = self.parent.NumClasses <= 2, class_weights=self.parent.data.class_weights)
        if self.loss == dlLoss.kl_divergence:
            return self.kullback_Leibler_divergence_loss_function()
        else:
            raise # not supported

    def categorical_cross_entropy_loss_function(self, binary, class_weights):
        # side note: in retina-net paper alpha = 0.75 gives best result for cce
        if class_weights is None:
            class_weights = 0.5
        if binary:
            return focal_loss_binary(usedistmap=False, class_weights=class_weights, gamma=0)
        else:
            return focal_loss(usedistmap=False, class_weights=class_weights, gamma=0)

    def focal_Loss_function(self, binary, class_weights):
        if binary:
            return focal_loss_binary(usedistmap=False, class_weights=class_weights, gamma=2)
        else:
            return focal_loss(usedistmap=False, class_weights=class_weights, gamma=2)
    
    def kullback_Leibler_divergence_loss_function(self):
        return kullback_Leibler_divergence_loss_weighted(False)




from .base import Metric
from .base import functional as F

SMOOTH = 1e-5


class IOUScore(Metric):
    r""" The `Jaccard index`_, also known as Intersection over Union and the Jaccard similarity coefficient
    (originally coined coefficient de communauté by Paul Jaccard), is a statistic used for comparing the
    similarity and diversity of sample sets. The Jaccard coefficient measures similarity between finite sample sets,
    and is defined as the size of the intersection divided by the size of the union of the sample sets:

    .. math:: J(A, B) = \frac{A \cap B}{A \cup B}

    Args:
        class_weights: 1. or ``np.array`` of class weights (``len(weights) = num_classes``).
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.
        smooth: value to avoid division by zero
        per_image: if ``True``, metric is calculated as mean over images in batch (B),
            else over whole batch
        threshold: value to round predictions (use ``>`` comparison), if ``None`` prediction will not be round

    Returns:
       A callable ``iou_score`` instance. Can be used in ``model.compile(...)`` function.

    .. _`Jaccard index`: https://en.wikipedia.org/wiki/Jaccard_index

    Example:

    .. code:: python

        metric = IOUScore()
        model.compile('SGD', loss=loss, metrics=[metric])
    """

    def __init__(
            self,
            class_weights=None,
            class_indexes=None,
            threshold=None,
            per_image=False,
            smooth=SMOOTH,
            name=None,
    ):
        name = name or 'iou_score'
        super().__init__(name=name)
        self.class_weights = class_weights if class_weights is not None else 1
        self.class_indexes = class_indexes
        self.threshold = threshold
        self.per_image = per_image
        self.smooth = smooth

    def __call__(self, gt, pr):
        return F.iou_score(
            gt,
            pr,
            class_weights=self.class_weights,
            class_indexes=self.class_indexes,
            smooth=self.smooth,
            per_image=self.per_image,
            threshold=self.threshold,
            **self.submodules
        )


class FScore(Metric):
    r"""The F-score (Dice coefficient) can be interpreted as a weighted average of the precision and recall,
    where an F-score reaches its best value at 1 and worst score at 0.
    The relative contribution of ``precision`` and ``recall`` to the F1-score are equal.
    The formula for the F score is:

    .. math:: F_\beta(precision, recall) = (1 + \beta^2) \frac{precision \cdot recall}
        {\beta^2 \cdot precision + recall}

    The formula in terms of *Type I* and *Type II* errors:

    .. math:: L(tp, fp, fn) = \frac{(1 + \beta^2) \cdot tp} {(1 + \beta^2) \cdot fp + \beta^2 \cdot fn + fp}

    where:
         - tp - true positives;
         - fp - false positives;
         - fn - false negatives;

    Args:
        beta: Integer of float f-score coefficient to balance precision and recall.
        class_weights: 1. or ``np.array`` of class weights (``len(weights) = num_classes``)
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.
        smooth: Float value to avoid division by zero.
        per_image: If ``True``, metric is calculated as mean over images in batch (B),
            else over whole batch.
        threshold: Float value to round predictions (use ``>`` comparison), if ``None`` prediction will not be round.
        name: Optional string, if ``None`` default ``f{beta}-score`` name is used.

    Returns:
        A callable ``f_score`` instance. Can be used in ``model.compile(...)`` function.

    Example:

    .. code:: python

        metric = FScore()
        model.compile('SGD', loss=loss, metrics=[metric])
    """

    def __init__(
            self,
            beta=1,
            class_weights=None,
            class_indexes=None,
            threshold=None,
            per_image=False,
            smooth=SMOOTH,
            name=None,
    ):
        name = name or 'f{}-score'.format(beta)
        super().__init__(name=name)
        self.beta = beta
        self.class_weights = class_weights if class_weights is not None else 1
        self.class_indexes = class_indexes
        self.threshold = threshold
        self.per_image = per_image
        self.smooth = smooth

    def __call__(self, gt, pr):
        return F.f_score(
            gt,
            pr,
            beta=self.beta,
            class_weights=self.class_weights,
            class_indexes=self.class_indexes,
            smooth=self.smooth,
            per_image=self.per_image,
            threshold=self.threshold,
            **self.submodules
        )


class Precision(Metric):
    r"""Creates a criterion that measures the Precision between the
    ground truth (gt) and the prediction (pr).

    .. math:: F_\beta(tp, fp) = \frac{tp} {(tp + fp)}

    where:
         - tp - true positives;
         - fp - false positives;

    Args:
        class_weights: 1. or ``np.array`` of class weights (``len(weights) = num_classes``).
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.
        smooth: Float value to avoid division by zero.
        per_image: If ``True``, metric is calculated as mean over images in batch (B),
            else over whole batch.
        threshold: Float value to round predictions (use ``>`` comparison), if ``None`` prediction will not be round.
        name: Optional string, if ``None`` default ``precision`` name is used.

    Returns:
        A callable ``precision`` instance. Can be used in ``model.compile(...)`` function.

    Example:

    .. code:: python

        metric = Precision()
        model.compile('SGD', loss=loss, metrics=[metric])
    """

    def __init__(
            self,
            class_weights=None,
            class_indexes=None,
            threshold=None,
            per_image=False,
            smooth=SMOOTH,
            name=None,
    ):
        name = name or 'precision'
        super().__init__(name=name)
        self.class_weights = class_weights if class_weights is not None else 1
        self.class_indexes = class_indexes
        self.threshold = threshold
        self.per_image = per_image
        self.smooth = smooth

    def __call__(self, gt, pr):
        return F.precision(
            gt,
            pr,
            class_weights=self.class_weights,
            class_indexes=self.class_indexes,
            smooth=self.smooth,
            per_image=self.per_image,
            threshold=self.threshold,
            **self.submodules
        )


class Recall(Metric):
    r"""Creates a criterion that measures the Precision between the
    ground truth (gt) and the prediction (pr).

    .. math:: F_\beta(tp, fn) = \frac{tp} {(tp + fn)}

    where:
         - tp - true positives;
         - fn - false negatives;

    Args:
        class_weights: 1. or ``np.array`` of class weights (``len(weights) = num_classes``).
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.
        smooth: Float value to avoid division by zero.
        per_image: If ``True``, metric is calculated as mean over images in batch (B),
            else over whole batch.
        threshold: Float value to round predictions (use ``>`` comparison), if ``None`` prediction will not be round.
        name: Optional string, if ``None`` default ``recall`` name is used.

    Returns:
        A callable ``recall`` instance. Can be used in ``model.compile(...)`` function.

    Example:

    .. code:: python

        metric = Precision()
        model.compile('SGD', loss=loss, metrics=[metric])
    """

    def __init__(
            self,
            class_weights=None,
            class_indexes=None,
            threshold=None,
            per_image=False,
            smooth=SMOOTH,
            name=None,
    ):
        name = name or 'recall'
        super().__init__(name=name)
        self.class_weights = class_weights if class_weights is not None else 1
        self.class_indexes = class_indexes
        self.threshold = threshold
        self.per_image = per_image
        self.smooth = smooth

    def __call__(self, gt, pr):
        return F.recall(
            gt,
            pr,
            class_weights=self.class_weights,
            class_indexes=self.class_indexes,
            smooth=self.smooth,
            per_image=self.per_image,
            threshold=self.threshold,
            **self.submodules
        )


# aliases
iou_score = IOUScore()
f1_score = FScore(beta=1)
f2_score = FScore(beta=2)
precision = Precision()
recall = Recall()



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 15:45:19 2020

@author: Koerber
"""

import tensorflow.keras.backend as K
from tensorflow.keras.metrics import MeanIoU
from tensorflow.keras.utils import get_custom_objects
import tensorflow as tf

# classification metrcis
def accuracy():
    return tf.keras.metrics.CategoricalAccuracy()

def accuracy_binary():
    return tf.keras.metrics.BinaryAccuracy()

def recall(y_true, y_pred):
    y_true = K.ones_like(y_true) 
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    all_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
    recall = true_positives / (all_positives + K.epsilon())
    return recall

def precision(y_true, y_pred):
    y_true = K.ones_like(y_true) 
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
    precision = true_positives / (predicted_positives + K.epsilon())
    return precision

def f1_score(y_true, y_pred):
    precision = precision_m(y_true, y_pred)
    recall = recall_m(y_true, y_pred)
    return 2*((precision*recall)/(precision+recall+K.epsilon()))


# regression         
def mean_squared_error():
    return tf.keras.metrics.MeanSquaredError()

def mean_absolute_error():
    return tf.keras.metrics.MeanAbsoluteError()

def mean_squared_logarithmic_error():
    return tf.keras.metrics.MeanSquaredLogarithmicError()

def root_mean_squared_error():
    return tf.keras.metrics.RootMeanSquaredError ()


# segmentation
def mean_iou_binary(y_true, y_pred):
    y_pred = K.cast(K.greater(y_pred, .5), dtype='float32')
    intersection = K.sum(y_true * y_pred)
    union = K.sum(y_true) + K.sum(y_pred) - intersection
    return K.mean((intersection + K.epsilon()) / (union + K.epsilon()))

def mean_iou_binary_weighted(y_true, y_pred):
    y_true = y_true[...,:-1]
    return mean_iou_binary(y_true, y_pred)

def mean_iou(y_true, y_pred):
    nb_classes = K.int_shape(y_pred)[-1]
    true_pixels = K.argmax(y_true, axis=-1) 
    pred_pixels = K.argmax(y_pred, axis=-1)
    iou = []
    for i in range(nb_classes-1):
        true_labels = K.cast(K.equal(true_pixels, i), K.floatx())
        pred_labels = K.cast(K.equal(pred_pixels, i), K.floatx())
        intersection = K.sum(true_labels * pred_labels)
        union = K.sum(true_labels) + K.sum(pred_labels) - intersection
        res = (intersection + K.epsilon()) / (union + K.epsilon())
        iou.append(res)
    return K.mean(tf.stack(iou))


def mean_iou_weighted(y_true, y_pred):
    y_true = y_true[...,:-1]
    return mean_iou(y_true, y_pred)


def dice_coefficient(y_true, y_pred):
    nb_classes = K.int_shape(y_pred)[-1]
    true_pixels = K.argmax(y_true, axis=-1) 
    pred_pixels = K.argmax(y_pred, axis=-1)
    iou = []
    for i in range(nb_classes-1):
        true_labels = K.cast(K.equal(true_pixels, i), K.floatx())
        pred_labels = K.cast(K.equal(pred_pixels, i), K.floatx())
        nom = 2.*K.sum(true_labels * pred_labels)
        denom = K.sum(true_labels) + K.sum(pred_labels)
        res = (nom + K.epsilon()) / (denom + K.epsilon())
        iou.append(res)
    return K.mean(tf.stack(iou))

def dice_coefficient_binary(y_true, y_pred):
    y_pred = K.cast(K.greater(y_pred, .5), dtype='float32')
    nom = 2.*K.sum(y_true * y_pred)
    denom = K.sum(y_true) + K.sum(y_pred)
    return K.mean((nom + K.epsilon()) / (denom + K.epsilon()))

def dice_coefficient_weighted(y_true, y_pred):
    y_true = y_true[...,:-1]
    return dice_coefficient(y_true, y_pred)

def dice_coefficient_binary_weighted(y_true, y_pred):
    y_true = y_true[...,:-1]
    return dice_coefficient_binary(y_true, y_pred)

def pixel_accuracy_binary(y_true, y_pred):
    y_pred = K.cast(K.greater(y_pred, .5), dtype='float32')
    return K.mean(tf.math.count_nonzero(y_pred == y_true))

def pixel_accuracy_binary_weighted(y_true, y_pred):
    y_true = y_true[...,:-1]
    return pixel_accuracy_binary(y_true, y_pred)

def pixel_accuracy(y_true, y_pred):
    true_pixels = K.argmax(y_true, axis=-1) 
    pred_pixels = K.argmax(y_pred, axis=-1)
    return K.mean(tf.math.count_nonzero(true_pixels == pred_pixels))

def pixel_accuracy_weighted(y_true, y_pred):
    y_true = y_true[...,:-1]
    return pixel_accuracy(y_true, y_pred)


# for loading models trained with custom metrics
get_custom_objects()['recall'] = recall
get_custom_objects()['precision'] = precision
get_custom_objects()['f1_score'] = f1_score



get_custom_objects()['pixel_accuracy_binary'] = pixel_accuracy_binary
get_custom_objects()['pixel_accuracy_binary_weighted'] = pixel_accuracy_binary_weighted
get_custom_objects()['pixel_accuracy'] = pixel_accuracy
get_custom_objects()['pixel_accuracy_weighted'] = pixel_accuracy_weighted

get_custom_objects()['mean_iou'] = mean_iou
get_custom_objects()['mean_iou_weighted'] = mean_iou_weighted
get_custom_objects()['mean_iou_binary'] = mean_iou_binary
get_custom_objects()['mean_iou_binary_weighted'] = mean_iou_binary_weighted

get_custom_objects()['dice_coefficient'] = dice_coefficient
get_custom_objects()['dice_coefficient_binary'] = dice_coefficient_binary
get_custom_objects()['dice_coefficient_weighted'] = dice_coefficient_weighted
get_custom_objects()['dice_coefficient_binary_weighted'] = dice_coefficient_binary_weighted


from dl.models.keras_applications import get_submodules_from_kwargs

from ._common_blocks import Conv2dBn
from ._utils import freeze_model
from ..backbones.backbones_factory import Backbones

backend = None
layers = None
models = None
keras_utils = None


# ---------------------------------------------------------------------
#  Utility functions
# ---------------------------------------------------------------------

def get_submodules():
    return {
        'backend': backend,
        'models': models,
        'layers': layers,
        'utils': keras_utils,
    }


# ---------------------------------------------------------------------
#  Blocks
# ---------------------------------------------------------------------

def Conv3x3BnReLU(filters, use_batchnorm, name=None):
    kwargs = get_submodules()

    def wrapper(input_tensor):
        return Conv2dBn(
            filters,
            kernel_size=3,
            activation='relu',
            kernel_initializer='he_uniform',
            padding='same',
            use_batchnorm=use_batchnorm,
            name=name,
            **kwargs
        )(input_tensor)

    return wrapper


def DoubleConv3x3BnReLU(filters, use_batchnorm, name=None):
    name1, name2 = None, None
    if name is not None:
        name1 = name + 'a'
        name2 = name + 'b'

    def wrapper(input_tensor):
        x = Conv3x3BnReLU(filters, use_batchnorm, name=name1)(input_tensor)
        x = Conv3x3BnReLU(filters, use_batchnorm, name=name2)(x)
        return x

    return wrapper


def FPNBlock(pyramid_filters, stage):
    conv0_name = 'fpn_stage_p{}_pre_conv'.format(stage)
    conv1_name = 'fpn_stage_p{}_conv'.format(stage)
    add_name = 'fpn_stage_p{}_add'.format(stage)
    up_name = 'fpn_stage_p{}_upsampling'.format(stage)

    channels_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    def wrapper(input_tensor, skip):
        # if input tensor channels not equal to pyramid channels
        # we will not be able to sum input tensor and skip
        # so add extra conv layer to transform it
        input_filters = backend.int_shape(input_tensor)[channels_axis]
        if input_filters != pyramid_filters:
            input_tensor = layers.Conv2D(
                filters=pyramid_filters,
                kernel_size=(1, 1),
                kernel_initializer='he_uniform',
                name=conv0_name,
            )(input_tensor)

        skip = layers.Conv2D(
            filters=pyramid_filters,
            kernel_size=(1, 1),
            kernel_initializer='he_uniform',
            name=conv1_name,
        )(skip)

        x = layers.UpSampling2D((2, 2), name=up_name)(input_tensor)
        x = layers.Add(name=add_name)([x, skip])

        return x

    return wrapper


# ---------------------------------------------------------------------
#  FPN Decoder
# ---------------------------------------------------------------------

def build_fpn(
        backbone,
        skip_connection_layers,
        pyramid_filters=256,
        segmentation_filters=128,
        classes=1,
        activation='sigmoid',
        use_batchnorm=True,
        aggregation='sum',
        dropout=None,
):
    input_ = backbone.input
    x = backbone.output

    # building decoder blocks with skip connections
    skips = ([backbone.get_layer(name=i).output if isinstance(i, str)
              else backbone.get_layer(index=i).output for i in skip_connection_layers])

    # build FPN pyramid
    p5 = FPNBlock(pyramid_filters, stage=5)(x, skips[0])
    p4 = FPNBlock(pyramid_filters, stage=4)(p5, skips[1])
    p3 = FPNBlock(pyramid_filters, stage=3)(p4, skips[2])
    p2 = FPNBlock(pyramid_filters, stage=2)(p3, skips[3])

    # add segmentation head to each
    s5 = DoubleConv3x3BnReLU(segmentation_filters, use_batchnorm, name='segm_stage5')(p5)
    s4 = DoubleConv3x3BnReLU(segmentation_filters, use_batchnorm, name='segm_stage4')(p4)
    s3 = DoubleConv3x3BnReLU(segmentation_filters, use_batchnorm, name='segm_stage3')(p3)
    s2 = DoubleConv3x3BnReLU(segmentation_filters, use_batchnorm, name='segm_stage2')(p2)

    # upsampling to same resolution
    s5 = layers.UpSampling2D((8, 8), interpolation='nearest', name='upsampling_stage5')(s5)
    s4 = layers.UpSampling2D((4, 4), interpolation='nearest', name='upsampling_stage4')(s4)
    s3 = layers.UpSampling2D((2, 2), interpolation='nearest', name='upsampling_stage3')(s3)

    # aggregating results
    if aggregation == 'sum':
        x = layers.Add(name='aggregation_sum')([s2, s3, s4, s5])
    elif aggregation == 'concat':
        concat_axis = 3 if backend.image_data_format() == 'channels_last' else 1
        x = layers.Concatenate(axis=concat_axis, name='aggregation_concat')([s2, s3, s4, s5])
    else:
        raise ValueError('Aggregation parameter should be in ("sum", "concat"), '
                         'got {}'.format(aggregation))

    if dropout:
        x = layers.SpatialDropout2D(dropout, name='pyramid_dropout')(x)

    # final stage
    x = Conv3x3BnReLU(segmentation_filters, use_batchnorm, name='final_stage')(x)
    x = layers.UpSampling2D(size=(2, 2), interpolation='bilinear', name='final_upsampling')(x)

    # model head (define number of output classes)
    x = layers.Conv2D(
        filters=classes,
        kernel_size=(3, 3),
        padding='same',
        use_bias=True,
        kernel_initializer='glorot_uniform',
        name='head_conv',
    )(x)
    x = layers.Activation(activation, name=activation)(x)

    # create keras model instance
    model = models.Model(input_, x)

    return model


# ---------------------------------------------------------------------
#  FPN Model
# ---------------------------------------------------------------------

def FPN(
        backbone_name='vgg16',
        input_shape=(None, None, 3),
        classes=21,
        activation='softmax',
        weights=None,
        encoder_weights='imagenet',
        encoder_freeze=False,
        encoder_features='default',
        pyramid_block_filters=256,
        pyramid_use_batchnorm=True,
        pyramid_aggregation='concat',
        pyramid_dropout=None,
        **kwargs
):
    """FPN_ is a fully convolution neural network for image semantic segmentation

    Args:
        backbone_name: name of classification model (without last dense layers) used as feature
                extractor to build segmentation model.
        input_shape: shape of input data/image ``(H, W, C)``, in general
                case you do not need to set ``H`` and ``W`` shapes, just pass ``(None, None, C)`` to make your model be
                able to process images af any size, but ``H`` and ``W`` of input images should be divisible by factor ``32``.
        classes: a number of classes for output (output shape - ``(h, w, classes)``).
        weights: optional, path to model weights.
        activation: name of one of ``keras.activations`` for last model layer (e.g. ``sigmoid``, ``softmax``, ``linear``).
        encoder_weights: one of ``None`` (random initialization), ``imagenet`` (pre-training on ImageNet).
        encoder_freeze: if ``True`` set all layers of encoder (backbone model) as non-trainable.
        encoder_features: a list of layer numbers or names starting from top of the model.
                Each of these layers will be used to build features pyramid. If ``default`` is used
                layer names are taken from ``DEFAULT_FEATURE_PYRAMID_LAYERS``.
        pyramid_block_filters: a number of filters in Feature Pyramid Block of FPN_.
        pyramid_use_batchnorm: if ``True``, ``BatchNormalisation`` layer between ``Conv2D`` and ``Activation`` layers
                is used.
        pyramid_aggregation: one of 'sum' or 'concat'. The way to aggregate pyramid blocks.
        pyramid_dropout: spatial dropout rate for feature pyramid in range (0, 1).

    Returns:
        ``keras.models.Model``: **FPN**

    .. _FPN:
        http://presentations.cocodataset.org/COCO17-Stuff-FAIR.pdf

    """
    global backend, layers, models, keras_utils
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)

    backbone = Backbones.get_backbone(
        backbone_name,
        input_shape=input_shape,
        weights=encoder_weights,
        include_top=False,
        **kwargs,
    )

    if encoder_features == 'default':
        encoder_features = Backbones.get_feature_layers(backbone_name, n=4)

    model = build_fpn(
        backbone=backbone,
        skip_connection_layers=encoder_features,
        pyramid_filters=pyramid_block_filters,
        segmentation_filters=pyramid_block_filters // 2,
        use_batchnorm=pyramid_use_batchnorm,
        dropout=pyramid_dropout,
        activation=activation,
        classes=classes,
        aggregation=pyramid_aggregation,
    )

    # lock encoder weights for fine-tuning
    if encoder_freeze:
        freeze_model(backbone, **kwargs)

    # loading model weights
    if weights is not None:
        model.load_weights(weights)

    return model



# -*- coding: utf-8 -*-
"""
Created on Mon Jan 20 13:42:08 2020

@author: Koerber
"""

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from ui.style import styleForm
from utils.shapes.ImageLabel import loadImageLabel


class Window(object):
    def setupUi(self, Form):
        Form.setWindowTitle('Classification Settings') 

        styleForm(Form)
        self.centralWidget = QWidget(Form)

        self.vlayout = QVBoxLayout(self.centralWidget)
        self.vlayout.setContentsMargins(0, 0, 0, 0)
        self.vlayout.setSpacing(6)
        self.centralWidget.setLayout(self.vlayout)
        
        self.centralWidget.setFixedSize(self.vlayout.sizeHint())
        Form.setFixedSize(self.vlayout.sizeHint())

class ClassificationResultsWindow(QMainWindow, Window):
    def __init__(self, parent, parentwindow):
        super(ClassificationResultsWindow, self).__init__(parent)
        self.parent = parent
        self.parentwindow = parentwindow
        self.setupUi(self)
        self.funcs = []


    def saveClasses(self, writer, images, labels):
        header = ['image name'] + ['frame number'] + ['class'] + ['classname']
        writer.writerow(header)
        num = len(labels)
        self.parent.emitinitProgress(num)
    
        for image, label in zip(images,labels):
            name = self.parent.files.getFilenameFromPath(self.parent.files.convertIfStackPath(image),withfileextension=True)
            classlabel,_ = loadImageLabel(self.parent.files.convertIfStackPath(label))
            if self.parent.files.isStackLabel(label):
                frame = self.parent.files.getFrameNumber(label) + 1
            else:
                frame = 1
            classname = self.parent.classList.getClassName(classlabel.classlabel)
            row = [name] + [frame] + [classlabel.classlabel] + [classname]
            writer.writerow(row) 
    



from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
import sys
from ui.style import setStyle

import platform

if platform.system() == 'Windows':
	import ctypes
	appid = u'mia_application'
	ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(appid)

def start():

    app = QCoreApplication.instance()
    # app = QApplication.instance()
    if app is None:
        app = QApplication(sys.argv)

    setStyle(app)
    splash_pix = QPixmap('icons/loading.png')
    load_screen = QSplashScreen(splash_pix, Qt.WindowStaysOnTopHint)
    load_screen.setMask(splash_pix.mask())
    load_screen.show()

    from mia_gui import MIA_UI
    dcdui = MIA_UI()

    load_screen.hide()
#    sys.exit(app.exec_())
    app.exec_()




if __name__ == "__main__":
    start()





from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
import random

from ui.style import getBrightColor

class QAdaptiveDoubleSpinBox(QDoubleSpinBox):
    # implementation of QDoubleSpinBox with adaptive step size,
    # will be obsolete in future as QAbstractSpinBox::AdaptiveDecimalStepType is supported with qt 9.12 (not in python for now)
    def __init__(self, parent):
        super(QAdaptiveDoubleSpinBox, self).__init__(parent)
        self.epsilon = 1e-8
        
    def stepBy(self,steps):
        value = self.value()
        
        if steps < 0:
            self.setSingleStep(value/2)
        else:
            self.setSingleStep(value)

        QDoubleSpinBox.stepBy(self,steps)
        
    def textFromValue(self, value):
        return QLocale().toString(round(value, 6), 'f', QLocale.FloatingPointShortest)

class LabelledSpinBox(QWidget):
    def __init__(self,labeltext, parent):
        super(LabelledSpinBox, self).__init__(parent)
        
        layout = QHBoxLayout()
        self.SpinBox = QSpinBox(parent)
        self.SpinBox.lineEdit().setObjectName('')
        layout.addWidget(self.SpinBox)
        self.Label = QLabel(parent)
        self.Label.setText(labeltext)
        layout.addWidget(self.Label)
        self.setLayout(layout)
        
    def objectName(self):
        return self.SpinBox.objectName(name)
        
    def setObjectName(self, name):
        self.SpinBox.setObjectName(name)

    def setToolTip(self, tooltip):
        self.Label.setToolTip(tooltip)
        self.SpinBox.setToolTip(tooltip)
        

class LabelledDoubleSpinBox(QWidget):
    def __init__(self,labeltext, parent):
        super(LabelledDoubleSpinBox, self).__init__(parent)

        
        layout = QHBoxLayout()
        self.SpinBox = QDoubleSpinBox(parent)
        self.SpinBox.lineEdit().setObjectName('')
        layout.addWidget(self.SpinBox)
        self.Label = QLabel(parent)
        self.Label.setText(labeltext)
        layout.addWidget(self.Label)
        self.setLayout(layout)
        
    def objectName(self):
        return self.SpinBox.objectName(name)
        
    def setObjectName(self, name):
        self.SpinBox.setObjectName(name)

    def setToolTip(self, tooltip):
        self.Label.setToolTip(tooltip)
        self.SpinBox.setToolTip(tooltip)

class LabelledAdaptiveDoubleSpinBox(QWidget):
    def __init__(self,labeltext, parent):
        super(LabelledAdaptiveDoubleSpinBox, self).__init__(parent)
        
        layout = QHBoxLayout()
        self.SpinBox = QAdaptiveDoubleSpinBox(parent)
        self.SpinBox.lineEdit().setObjectName('')
        layout.addWidget(self.SpinBox)
        self.Label = QLabel(parent)
        self.Label.setText(labeltext)
        layout.addWidget(self.Label)
        self.setLayout(layout)
        
    def objectName(self):
        return self.SpinBox.objectName(name)
        
    def setObjectName(self, name):
        self.SpinBox.setObjectName(name)
        

    def setToolTip(self, tooltip):
        self.Label.setToolTip(tooltip)
        self.SpinBox.setToolTip(tooltip)

class DCDButton(QPushButton):
    isclicked = pyqtSignal()
    
    def __init__(self, parent, text=None):
        super(DCDButton, self).__init__(parent)
        self.parent = parent
        if text:
            self.setText(text)
            self.setStyleSheet('text-align:left')
        self.setStyleSheet('QPushButton:hover:!pressed { border: 1px solid lightblue}')
        self.setFlat(True)
        self.clicked.connect(self.on_clicked)
        
    def on_clicked(self):
        self.isclicked.emit()
        self.parent.setFocus()
        
        
    def resetStyle(self):
        self.setStyleSheet('QPushButton:hover:!pressed { border: 1px solid lightblue}')
        
class QSetting(QWidget):
    valueChanged = pyqtSignal() 
    def __init__(self, parent, name, value=None):
        super(QWidget, self).__init__(parent)
        self._value = value
        self.setObjectName(name)
        self.setVisible(False)
        
    def value(self):
        return self._value
    
    def setValue(self, v):
        self._value = v
        self.valueChanged.emit()


class ClassList(QListWidget):
    def __init__(self, parent):
        super(ClassList, self).__init__(parent)
        self.parent = parent
        self.setSelectionMode(QAbstractItemView.NoSelection)
        self.setStyleSheet("QListWidget { border: 2px solid " + getBrightColor(asString=True) +";}")

    def addClass(self, name = 'class type'):
        item = QListWidgetItem(self)
        self.addItem(item)
        id = self.row(item)
        row = ClassWidget(name, id , self.parent)
        # not so nice to hard code size here
        item.setSizeHint(row.minimumSizeHint() + QSize(100,0) )
        self.setItemWidget(item, row)
        self.resize()
        self.parent.ClassListUpdated()
        
    def removeClass(self):
        idx = self.count()-1
        if idx <= 1:
            return
        else:
            self.takeItem(idx)
            self.resize()
        if self.getActiveClass() == -1:
            self.setClass(0)    
        self.parent.ClassListUpdated()

    def resize(self):
        dynamic_height = min(self.sizeHintForRow(0) * self.count() + 2 * self.frameWidth(), 380)
        self.setFixedSize(self.sizeHintForColumn(0) + 2 * self.frameWidth(), dynamic_height)

    def setClass(self, i):
        item = self.item(i)
        if item:
            self.itemWidget(item).rb_select.setChecked(True)
            
    def setClassWeight(self, classnum, weight):
        item = self.item(classnum)
        if item:
            self.itemWidget(item).classweight.setValue(weight)
            
    def getClassWeight(self, classnum):
        item = self.item(classnum)
        if item:
            return self.itemWidget(item).classweight.value()
            
    def getClassName(self, i):
        item = self.item(i)
        if item:
            return self.itemWidget(item).edit_name.text()
        
    def getNumberOfClasses(self):
        return self.count()
    
    def setNumberOfClasses(self, numclasses):
        diff = numclasses - self.getNumberOfClasses()
        if diff > 0:
            for i in range(diff):
                self.addClass()
        if diff < 0:
            for i in range(abs(diff)):
                self.removeClass()
            

    def getActiveClass(self):
        for i in range(self.count()):
            if self.itemWidget(self.item(i)).rb_select.isChecked():
                return i
        return -1
    
    def getClassColor(self, c):
        while c >= self.count() and c < 255:
            self.addClass()
        return self.itemWidget(self.item(c)).color
        

        
class ClassWidget(QWidget):
    ButtonGroup = QButtonGroup()
    def __init__(self, name, id, parent):
        super(ClassWidget, self).__init__(parent)
        self.id = id
        self.parent = parent
        r = random.randint(0,255)
        g = random.randint(0,255)
        b = random.randint(0,255)
        self.color = QColor(r,g,b)
        
        self.row = QHBoxLayout()
        self.rb_select = QRadioButton(parent)  
        self.rb_select.setToolTip('Select class')
        self.ButtonGroup.addButton(self.rb_select)

        self.pb_color = DCDButton(parent)
        self.pb_color.setToolTip('Change color')
        self.pb_color.setMaximumSize(15, 15)
        self.setButtonColor(self.color)
        
        self.edit_name = QLineEdit(parent)
        self.edit_name.setText(name)
        self.edit_name.setToolTip('Edit class name')
        self.edit_name.setObjectName('nn_class_%i' % self.id)
        self.edit_name.setStyleSheet("QLineEdit { border: None }")

        self.classweight = QSetting(parent, 'nn_classweight_%i' % self.id, 0.5)
        
        
        self.label_numOfContours = QLabel(parent)
        self.label_numOfContours.setObjectName('numOfContours_class' + str(self.id))
        self.label_numOfContours.setToolTip('Number of contours')
        self.label_numOfContours.setText('0')     
        
        self.row.addWidget(self.rb_select)
        self.row.addWidget(self.pb_color)
        self.row.addWidget(self.edit_name)
        self.row.addWidget(self.label_numOfContours)
        self.setLayout(self.row)
        
        self.pb_color.clicked.connect(self.changeColor)
        self.rb_select.toggled.connect(self.selectClass)
        self.edit_name.textChanged.connect(self.nameChanged)

    
    def setButtonColor(self, color):
        self.color = color
        pixmap = QPixmap(15, 15)
        pixmap.fill(color)
        self.pb_color.setIcon(QIcon(pixmap))
 
        
    def changeColor(self):
        color = QColorDialog.getColor()
        if color.isValid():
            self.setButtonColor(color)
            self.parent.classcolorChanged()

    def selectClass(self):
        self.parent.classChanged()
        
    def nameChanged(self):
        self.parent.ClassListUpdated()


# open dialogs
def openFolder(header):
    dialog = QFileDialog()
    dialog.setWindowTitle(header)
    dialog.setFileMode(QFileDialog.DirectoryOnly)
    dialog.setOption(QFileDialog.DontUseNativeDialog, True)
    dialog.setOption(QFileDialog.ShowDirsOnly, False)
    
    if dialog.exec() == QDialog.Accepted:
        return dialog.selectedFiles()[0]
    else:
        return False

def saveFile(header, filter, suffix, defaultname = None):
    dialog = QFileDialog()
    dialog.setWindowTitle(header)
    dialog.setOption(QFileDialog.DontUseNativeDialog, True)
    dialog.setDefaultSuffix(suffix)
    dialog.setAcceptMode(QFileDialog.AcceptSave)
    if defaultname:
        dialog.selectFile(defaultname)
    dialog.setNameFilters([filter])
    if dialog.exec_() == QDialog.Accepted:
        return dialog.selectedFiles()[0]
    else:
       return False

def loadFile(header, filter):
    dialog = QFileDialog()
    dialog.setWindowTitle(header)
    dialog.setOption(QFileDialog.DontUseNativeDialog, True)
    dialog.setAcceptMode(QFileDialog.AcceptOpen)
    dialog.setNameFilters([filter])
    if dialog.exec_() == QDialog.Accepted:
        return dialog.selectedFiles()[0]
    else:
       return False



# this is a clone from https://github.com/qubvel/segmentation_models
# with minor changes
# licence MIT, https://github.com/qubvel/segmentation_models/blob/master/LICENSE


import os
import functools
from .__version__ import __version__
from . import base

_KERAS_FRAMEWORK_NAME = 'keras'
_TF_KERAS_FRAMEWORK_NAME = 'tf.keras'

_DEFAULT_KERAS_FRAMEWORK = _KERAS_FRAMEWORK_NAME
_KERAS_FRAMEWORK = None
_KERAS_BACKEND = None
_KERAS_LAYERS = None
_KERAS_MODELS = None
_KERAS_UTILS = None
_KERAS_LOSSES = None


def inject_global_losses(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        kwargs['losses'] = _KERAS_LOSSES
        return func(*args, **kwargs)

    return wrapper


def inject_global_submodules(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        kwargs['backend'] = _KERAS_BACKEND
        kwargs['layers'] = _KERAS_LAYERS
        kwargs['models'] = _KERAS_MODELS
        kwargs['utils'] = _KERAS_UTILS
        return func(*args, **kwargs)

    return wrapper


def filter_kwargs(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        new_kwargs = {k: v for k, v in kwargs.items() if k in ['backend', 'layers', 'models', 'utils']}
        return func(*args, **new_kwargs)

    return wrapper


def framework():
    """Return name of Segmentation Models framework"""
    return _KERAS_FRAMEWORK


def set_framework(name):
    """Set framework for Segmentation Models

    Args:
        name (str): one of ``keras``, ``tf.keras``, case insensitive.

    Raises:
        ValueError: in case of incorrect framework name.
        ImportError: in case framework is not installed.

    """
    name = name.lower()

    if name == _KERAS_FRAMEWORK_NAME:
        import keras
        import dl.models.efficientnet.keras  # init custom objects
    elif name == _TF_KERAS_FRAMEWORK_NAME:
        from tensorflow import keras
        import dl.models.efficientnet.tfkeras  # init custom objects
    else:
        raise ValueError('Not correct module name `{}`, use `{}` or `{}`'.format(
            name, _KERAS_FRAMEWORK_NAME, _TF_KERAS_FRAMEWORK_NAME))

    global _KERAS_BACKEND, _KERAS_LAYERS, _KERAS_MODELS
    global _KERAS_UTILS, _KERAS_LOSSES, _KERAS_FRAMEWORK

    _KERAS_FRAMEWORK = name
    _KERAS_BACKEND = keras.backend
    _KERAS_LAYERS = keras.layers
    _KERAS_MODELS = keras.models
    _KERAS_UTILS = keras.utils
    _KERAS_LOSSES = keras.losses

    # allow losses/metrics get keras submodules
    base.KerasObject.set_submodules(
        backend=keras.backend,
        layers=keras.layers,
        models=keras.models,
        utils=keras.utils,
    )


# set default framework
_framework = os.environ.get('SM_FRAMEWORK', _DEFAULT_KERAS_FRAMEWORK)
try:
    set_framework(_framework)
except ImportError:
    other = _TF_KERAS_FRAMEWORK_NAME if _framework == _KERAS_FRAMEWORK_NAME else _KERAS_FRAMEWORK_NAME
    set_framework(other)

# import helper modules
#from . import losses
#from . import metrics
from . import utils

# wrap segmentation models with framework modules
from .backbones.backbones_factory import Backbones
from .models.unet import Unet as _Unet
from .models.pspnet import PSPNet as _PSPNet
from .models.linknet import Linknet as _Linknet
from .models.fpn import FPN as _FPN

Unet = inject_global_submodules(_Unet)
PSPNet = inject_global_submodules(_PSPNet)
Linknet = inject_global_submodules(_Linknet)
FPN = inject_global_submodules(_FPN)
get_available_backbone_names = Backbones.models_names


def get_preprocessing(name):
    preprocess_input = Backbones.get_preprocessing(name)
    # add bakcend, models, layers, utils submodules in kwargs
    preprocess_input = inject_global_submodules(preprocess_input)
    # delete other kwargs
    # keras-applications preprocessing raise an error if something
    # except `backend`, `layers`, `models`, `utils` passed in kwargs
    preprocess_input = filter_kwargs(preprocess_input)
    return preprocess_input


__all__ = [
    'Unet', 'PSPNet', 'FPN', 'Linknet',
    'set_framework', 'framework',
    'get_preprocessing', 'get_available_backbone_names',
    'losses', 'metrics', 'utils',
    '__version__',
]



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 22 14:23:06 2019

@author: Koerber
"""

import cv2
import numpy as np

import math

from skimage import morphology
from skimage.morphology import medial_axis, skeletonize

    

BACKGROUNDCLASS = 255
from utils.shapes.Shape import Shape, Shapes, FreeFormContour_ID, FreeFormContour_ID_legacy


class Contours(Shapes):
    def __init__(self):
        self.minSize = 3
        super(Contours,self).__init__(FreeFormContour_ID)
        
    def ShapeAlreadyExist(self,c):
        return checkIfContourInListOfContours(c, self.shapes) and c.isValid(self.minSize)
            
    def addShapes(self,cts):
        cnts = getContoursNotinListOfContours(cts, self.shapes)
        new_cnts = [ x for x in cnts if x.labeltype == self.labeltype and x.isValid(self.minSize)]
        self.shapes.extend(new_cnts)

    def deleteShapes(self, contours):
        for c in contours:
            self.deleteShape(c)

    def load(self, filename):
        cnts = loadContours(filename)
        self.shapes = [ x for x in cnts if x.isValid(self.minSize)]
        
    def save(self, filename):
        saveContours(self.shapes, filename)    

            
    def getclosestContour(self, point, dist, label = None):
        for c in self.shapes: 
            if -c.distance(point) < dist:
                if label is None:
                    return c
                elif label == c.classlabel:
                    return c
        return None


# freeform
class Contour(Shape):
    def __init__(self, classlabel, points = None, inner = None):
        super(Contour,self).__init__(classlabel, FreeFormContour_ID)
        self.points = None
        self.boundingbox = None # for performance issues saved as member var
        self.moments = None
        self.skeleton = None
        self.innercontours = []
        self.innerparams = []
        

        if isinstance(points, np.ndarray):
            if len(points.shape) == 2:
                self.points = [points]
            else:
                self.points = points
                if isinstance(inner, np.ndarray):
                    self.unpackInnerContours(inner)
                
        elif points is None:
            self.points = None

    def __del__(self): 
        self.innercontours.clear()

    def setClassLabel(self,n):
        self.classlabel = n
            
    def addPoint(self, point):
        self.points = np.concatenate([self.points, [point]])
        self.boundingbox = None
            
    def closeContour(self):
        self.points = np.concatenate([self.points, self.points[0].reshape(1,1,2)])
        
    def getFirstPoint(self):
        return self.points[0][0]
    
    def getLastPoint(self):
        return self.points[-1][0]

    def isValid(self, minArea = 3):
        if len(self.points) >= 3 and cv2.contourArea(self.points) > minArea:   
            return True
        else:
            return False

    def addInnerContour(self, c):
        if len(c) >= 3 and cv2.contourArea(c) > 3:
            self.innercontours.append(c)
            
    def packedInnerContours(self):
        if len(self.innercontours) < 1:
            return self.innercontours
        else:
            # we pack a [-1,-1] at front of every contour
            # warning: bracket overload!!! :)          
            return np.concatenate([np.concatenate((np.array([[[-1,-1]]]),x)) for x in self.innercontours])
        
    def unpackInnerContours(self, inner_packed):
        if len(inner_packed) == 0:
            return
        # we split the array at the position of every [-1,-1] into separate contours and remove [-1,-1]
        separate = np.where(inner_packed == ([-1,-1]))
        s = separate[0][0::2]
        inner = np.split(inner_packed, s[1:])    
        self.innercontours = [x[1:,...] for x in inner]


    def getInnerContourParams(self):
        if not self.innerparams:
            self.innerparams = [ (cv2.moments(x),cv2.boundingRect(x)) for x in self.innercontours]
        return self.innerparams

    def getBoundingBox(self):
        if self.boundingbox is None:
            if self.numPoints() == 1:
                return self.points[0][0][0], self.points[0][0][1], 0,0
            else:
                self.boundingbox = cv2.boundingRect(self.points)
        return self.boundingbox
    
    def getMoments(self):
        if self.moments is None:
            self.moments = cv2.moments(self.points)
        return self.moments
        
    def getSize(self):
        # careful: contourarea uses Green formula
        # as we use this in segmentation non_zero pixels might be more accurate
        # consider changing to non_zero pixels
        outer = cv2.contourArea(self.points)
        inner = [cv2.contourArea(x) for x in self.innercontours]
        return outer - sum(inner)
    
    def numPoints(self):
        return len(self.points)
                    
    def inside(self, point, distance=0):
        # distance ignored atm, if needed use cv2.pointPolygonTest(self.points,  p, True) 
        p = (point[0,0], point[0,1])
        inside = cv2.pointPolygonTest(self.points,  p, False)   
        if inside < 0:
            return False
        if not self.innercontours:
            return True    
        else:
            return next((False for c in self.innercontours if cv2.pointPolygonTest(c,  p, False) > 0), True)

    def getBottomPoint(self):
        b = (self.points[self.points[..., 1].argmax()][0])
        return (b[0], b[1])
    
    def getTopPoint(self):
        b = (self.points[self.points[..., 1].argmin()][0])
        return (b[0], b[1])
    
    def getLeftPoint(self):
        b = (self.points[self.points[..., 0].argmin()][0])
        return (b[0], b[1])
    
    def getRightpPoint(self):
        b = (self.points[self.points[..., 0].argmax()][0])
        return (b[0], b[1])
    
    def distance(self, point):
        return cv2.pointPolygonTest(self.points,  (point[0,0], point[0,1]), True) 

    def getPosition(self):
        return self.getCenter()
    
    def getCenter(self):
        M = cv2.moments(self.points)
        if M["m00"] == 0:
            raise Exception('division by zero')
        cX = int(M["m10"] / M["m00"])
        cY = int(M["m01"] / M["m00"])
        return (cX, cY)
    
    def getPerimeter(self):
        return cv2.arcLength(self.points,True)
    
    def getSkeletonLength(self, smoothing):
        if self.skeleton is None:
            self.skeleton = self.getSkeleton(smoothing)
        if self.skeleton is None:
            return 0
        else:
            return sum([cv2.arcLength(c, True) for c in self.skeleton])/2
#            return cv2.arcLength(self.skeleton, False)/2
    
    def getSkeleton(self, smoothing):
        if self.skeleton is not None:
            return self.skeleton
        t = self.points[self.points[..., 1].argmin()][0][1]
        b = self.points[self.points[..., 1].argmax()][0][1]
        l = self.points[self.points[..., 0].argmin()][0][0]
        r = self.points[self.points[..., 0].argmax()][0][0]

        width = r-l
        height = b-t

        if width < 3 or height < 3:
            return None

        image = np.zeros((height, width), np.uint8)
#        image = np.zeros((height, width, 1), np.uint8)

        cv2.drawContours(image, [self.points - (l,t)], 0, (255), -1)
        inner = [i - (l,t) for i in self.innercontours]
        cv2.drawContours(image, inner, -1, (0), -1)
        
        
        blurred = cv2.medianBlur(image, smoothing)

        skel = morphology.skeletonize(blurred>0,  method='lee')

        # contour based length measurement
        contours, _ = findContours(skel)

        self.skeleton = [cv2.approxPolyDP(c + ([l,t]), 1, True) for c in contours if len(c)>0]
        return self.skeleton
    
        ## pixel based length measurement
#        diag_kernel = np.array(([-1, 1, -1],	[1, -10, 1],	[-1, 1, -1]), dtype="int")
#        diac = cv2.filter2D(skel,-1, diag_kernel)
#        diag_steps = np.count_nonzero(diac)//2 - 1
#        total = np.count_nonzero(skel)
#        length = total - diag_steps + diag_steps * np.sqrt(2)
#        pts = np.where(skel>0) 
#        x = pts[0] + t
#        y = pts[1] + l
#
#        return x,y


# utilities
def drawcontour(image, contour, ignoreclasslabel=False, separateContours = False):
    cnt = contour.points
    ref_image = image.copy()
    if contour.numPoints() > 0:
        if separateContours:
            cv2.drawContours(image, [cnt], 0, (0), 3)  
            cv2.drawContours(image, [cnt], 0, (1), -1) 
        else:
            if ignoreclasslabel:
                cv2.drawContours(image, [cnt], 0, (1), -1)  
                cv2.drawContours(image, contour.innercontours, -1, (0), -1)
                # the contour of inner needs to be redrawn and belongs to outer, otherwise the inner contour is growing on each iteration
                cv2.drawContours(image, contour.innercontours, -1, (1), 1)
            elif contour.classlabel != 0:
                cv2.drawContours(image, [cnt], 0, (int(contour.classlabel)), -1)  
                cv2.drawContours(image, contour.innercontours, -1, (BACKGROUNDCLASS), -1)
                # the contour of inner needs to be redrawn and belongs to outer, otherwise the inner contour is growing on each iteration
                cv2.drawContours(image, contour.innercontours, -1, (int(contour.classlabel)), 1)
                image[image==BACKGROUNDCLASS] = ref_image[image==BACKGROUNDCLASS]
            else:
                cv2.drawContours(image, [cnt], 0, (BACKGROUNDCLASS), -1)  
                cv2.drawContours(image, contour.innercontours, -1, (0), -1)
                # the contour of inner needs to be redrawn and belongs to outer, otherwise the inner contour is growing on each iteration
                cv2.drawContours(image, contour.innercontours, -1, (BACKGROUNDCLASS), 1)


def drawContoursToLabel(label, contours, drawbackground = True):
    # split contours
    bg_contours, target_contours = [], []
    for c in contours:
       target = bg_contours if c.classlabel == 0 else target_contours
       target.append(c)
       
    if drawbackground:
        label = drawbackgroundToLabel(label, bg_contours)
      
    for c in target_contours:
        drawcontour(label, c)

    if not drawbackground:
        label[label==BACKGROUNDCLASS] = 0

    return label

def drawbackgroundToLabel(label, background):
    if not background or background == list():
        label[:] = (BACKGROUNDCLASS)
    else:
        for c in background:
            drawcontour(label, c) 
    return label
       
def extractContoursFromLabel(image, ext_only = False, offset=(0,0)):
    image = np.squeeze(image).astype(np.uint8)
    ret_contours = []

    # contours
    if np.all(image == BACKGROUNDCLASS):
        return []
    maxclass = np.max(image[image!=BACKGROUNDCLASS])
    for i in range(maxclass+1):
        if i == 0: # background
            if np.all(image != 0):
                continue
            else:
                thresh = (image == BACKGROUNDCLASS).astype(np.uint8)
        else:
            thresh = (image == i).astype(np.uint8)
        contours, hierarchy = findContours(thresh, ext_only, offset)
        if contours is not None:
            counter = -1
            for k, c in enumerate(contours):
                parent = hierarchy [0][k][3]
                if parent > -1:
                    ret_contours[counter].addInnerContour(c)
                else:
                    ret_contours.append(Contour(i,c))
                    counter += 1
    ret_contours.reverse()
    return ret_contours

def drawContoursToImage(image, contours, separate = False): 
    for c in contours:
        drawcontour(image, c, ignoreclasslabel=True, separateContours=separate)

def extractContoursFromImage(image, ext_only = False, offset = (0,0)):
    image = np.squeeze(image).astype(np.uint8)
    ret_contours = []
    contours, hierarchy = findContours(image, ext_only, offset)

    if contours is not None:
        counter = -1
        for k,c in enumerate(contours):
            parent = hierarchy [0][k][3]
            if parent > -1:
                ret_contours[counter].addInnerContour(c)
            else:
                ret_contours.append(Contour(-1,c))
                counter += 1
    ret_contours.reverse()
    return ret_contours

def packContours(contours):
    if not contours or len(contours) == 0:
        return None

    f1 = lambda x: np.array((x.classlabel, x.objectNumber),dtype=int)
    f2 = lambda x: x.points
#    f3 = lambda x: x.innercontours # we need to allow_pickle = True in np.load then
    f3 = lambda x: x.packedInnerContours()
    cnts = [f(x) for x in contours if x.isValid() for f in (f1, f2, f3)]
    cnts.insert(0,contours[0].labeltype)
    return cnts
    
def unpackContours(data):
    ret = []
    
    if data[0] == FreeFormContour_ID_legacy:
        label = data[1::2]
        array = data[2::2]
        ret = [Contour(x,y) for x,y in zip(label,array)]
    elif data[0] == FreeFormContour_ID:
        label = data[1::3]
        array = data[2::3]
        inner = data[3::3]
        ret = [Contour(x,y,z) for x,y,z in zip(label,array,inner)]
    return ret

def saveContours(contours, filename):
    cnts = packContours(contours)
    if cnts:
        np.savez(filename, *cnts)
        
def loadContours(filename):
    container = np.load(filename)
    data = [container[key] for key in container]
    return unpackContours(data)

def findContours(image, ext_only = False, offset=(0,0)):
    structure = cv2.RETR_EXTERNAL if ext_only else cv2.RETR_CCOMP

    # requires opencv 4 or greater, opencv 3 returns 3 params
    contours, hierarchy = cv2.findContours(image, structure , cv2.CHAIN_APPROX_SIMPLE, offset = offset)

    return contours, hierarchy

def removeBackground(image):
    image[image==BACKGROUNDCLASS] = 0
    return image

def checkIfContourInListOfContours(contour, contours):
    # returns True if contour is contained in contours
    return next((True for elem in contours if (elem.getMoments() == contour.getMoments() and elem.getBoundingBox() == contour.getBoundingBox() and elem.getInnerContourParams() == contour.getInnerContourParams())), False)

def getContoursNotinListOfContours(contours1, contours2):
    # returns all contours in contours1 not in contours2
    c1 = [(x, x.getMoments(), x.getBoundingBox(), x.getInnerContourParams() ) for x in contours1]
    c2 = [( x.getMoments(), x.getBoundingBox(), x.getInnerContourParams() ) for x in contours2]   
    return [x[0] for x in c1 if x[1:] not in c2]

def matchcontours(contour1, contour2):
    c1 = contour1.getCenter()
    c2 = contour2.getCenter()

    # potentially cv2.matchShapes(contour1.points, contour2.points, cv2.CONTOURS_MATCH_I1,0) based on Hu-moments
    # could be used additionally to compare, for now only the distance is calculated 
    return math.sqrt((c1[0]-c2[0])**2 + (c1[1]-c2[1])**2)

def getContourMinIntensity(image, contour):
    mask = np.zeros(image.shape[0:2], dtype = np.uint8)
    cv2.drawContours(mask, contour.points, -1,255,-1)
    minval = np.min(image[mask>0])
    return minval

def getContourMeanIntensity(image, contour):
    mask = np.zeros(image.shape[0:2], dtype = np.uint8)
    cv2.drawContours(mask, contour.points, -1,255,-1)
    meanval = np.mean(image[mask>0])
    return meanval

def getContourMaxIntensity(image, contour):
    mask = np.zeros(image.shape[0:2], dtype = np.uint8)
    cv2.drawContours(mask, contour.points, -1,255,-1)
    maxval = np.max(image[mask>0])
    return maxval


# -*- coding: utf-8 -*-
"""
Created on Fri Jun 26 15:37:52 2020

@author: Koerber
"""


from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *


from abc import ABC, abstractmethod
import numpy as np
import cv2
import utils.shapes.Contour as Contour
from ui.Tools import canvasTool
import random


from dl.data.labels import getMatchingImageLabelPairPaths

class Painter(ABC):
    # every painter can draw contours 
    def __init__(self, canvas):
        self.canvas = canvas
        
        self.contours = Contour.Contours()
        self.NewContour = None
        # self.sketch = None
        self.tools = []
        self.tools.append(canvasTool.drag)
        self.tools.append(canvasTool.draw)
        self.tools.append(canvasTool.extend)
        self.tools.append(canvasTool.poly)
        self.tools.append(canvasTool.delete)
        self.trackChanges = []
        self.maxChanges = 4
        self.contoursketch = None
        self.mask = None
        self.objectcolors = []
    
    def clear(self):
        self.contours.clear()
        self.resetSmartMask()
        
    def resetSmartMask(self):
        self.mask = None
        
    @abstractmethod
    def load(self):
        pass
        
    @abstractmethod
    def save(self):
        pass
    
    @abstractmethod
    def clearFOV(self):
        pass
    
    @abstractmethod
    def addShapes(self, shapes):
        pass
    
    @abstractmethod
    def checkForChanges(self):
        pass

    @abstractmethod
    def undo(self):
        pass

    def clearTracking(self):
        self.trackChanges.clear()

    def undoPossible(self):
        return len(self.trackChanges) >= 2

    @property
    def shapes(self):
        return self.contours
    
    @property
    def backgroundshapes(self):
        return self.contours.getShapesOfClass_x(0)
    
    @property
    def sketch(self):
        return self.contoursketch

    
    def draw(self):
        self.drawBackground()
    
    def getObjectColor(self, objnum):
        if len(self.objectcolors) < objnum:
            for i in range(0, objnum-len(self.objectcolors)):
                self.objectcolors.append(QColor(random.randint(0,255),random.randint(0,255),random.randint(0,255)))
        return self.objectcolors[objnum-1]

    def enableDrawBackgroundMode(self):
        self.tools.clear()
        self.tools.append(canvasTool.drag)
        self.tools.append(canvasTool.scale)
        self.tools.append(canvasTool.draw)
        self.tools.append(canvasTool.extend)
        self.tools.append(canvasTool.poly)
        self.tools.append(canvasTool.delete)

    def disableDrawBackgroundMode(self):
        self.tools.clear()
        self.tools.append(canvasTool.drag)
        self.tools.append(canvasTool.scale)
        
    def getPainter(self, color = None):
        if color is None:
            color = self.canvas.parent.ClassColor(self.canvas.parent.activeClass())
        # painter objects can only exist once per QWidget
        p = QPainter(self.canvas.image())
        color.setAlpha(255)    
        p.setPen(QPen(color, self.canvas.pen_size, Qt.SolidLine, Qt.SquareCap, Qt.BevelJoin))

        p.setBrush(QBrush(QColor(color),Qt.SolidPattern))
        p.setFont(QFont("Fixed",self.canvas.FontSize))
        return p
    
    def setColorTransparency(self, painter, color, transparency):
        if transparency == 1:        # opaque
            color.setAlpha(255)   
        elif transparency == -1:       # invisible
            color.setAlpha(0)
        elif transparency == 0:       # transparent
            color.setAlpha(self.canvas.ContourTransparency)
        painter.setBrush(QBrush(QColor(color),Qt.SolidPattern))
        

    def setPainterColor(self, painter, color):
        color.setAlpha(255)
        painter.setPen(QPen(color, self.canvas.pen_size, Qt.SolidLine, Qt.SquareCap, Qt.BevelJoin))
        painter.setBrush(QBrush(QColor(color),Qt.SolidPattern))

        
    def addline(self, p1 ,p2, color = None, dashed = False):
        p = self.getPainter(color)
        if dashed:
            pen = p.pen()
            pen.setStyle(Qt.DashDotDotLine)
            p.setPen(pen)
        else:
            pen = p.pen()
            pen.setStyle(Qt.SolidLine)
            p.setPen(pen)
        
        p.drawLine(p2, p1)
        self.canvas.updateImage()
        
    def addTriangle(self, point, size, color = None):
        psize = self.canvas.pen_size
        self.canvas.pen_size = 1
        p = self.getPainter(color)
        polygon = QPolygonF()
        polygon.append(point)
        point2 = QPointF(point.x() + size/2, point.y() - size)
        polygon.append(point2)
        point3 = QPointF(point.x() - size/2, point.y() - size)
        polygon.append(point3)
        p.drawPolygon(polygon)
        self.canvas.pen_size = psize
        self.canvas.updateImage()
        
    def addCircle(self, center, radius, color = None, erase = False):
        p = self.getPainter(color)
        p.setPen(QPen(Qt.NoPen))
        if erase:
            p.setBrush(QBrush(self.canvas.displayedimage.getImage()))
        p.drawEllipse(center, radius, radius)
        self.canvas.updateImage()


    def addRectangle(self, x, y, width, height, color = None):
        p = self.getPainter(color)
        p.drawRect(x, y, width, height)
        self.canvas.updateImage()


    #### contour drawing functions
    def drawcontour(self, contour):   
        painter = self.getPainter()
        path = self.contour2Path(contour)
        if not path.isEmpty():    
            if self.canvas.parent.colorCodeObjects() and contour.objectNumber != -1:
                color = self.getObjectColor(contour.objectNumber)
            else:
                color = self.canvas.parent.ClassColor(contour.classlabel)
            self.setPainterColor(painter, color)
            self.setColorTransparency(painter, color, transparency = 0)
            painter.drawPath(path)
            self.drawtrack(contour, painter, color)
            self.drawcontouraccessories(contour, painter)
            
            
    def drawBackground(self):    
        painter = self.getPainter()
        color = self.canvas.parent.ClassColor(0)
        self.setPainterColor(painter, color)
        bg = self.contours.getShapesOfClass_x(0)
        if len(bg) > 0:
            pm = self.canvas.displayedimage.getCanvas().copy()
            self.canvas.displayedimage.getCanvas().fill(QColor(0, 0, 0, self.canvas.ContourTransparency))
            painter.setCompositionMode(QPainter.CompositionMode_Source)
        
        for c in bg:  
            path = self.contour2Path(c)
            brush = QBrush(pm)             
            painter.fillPath(path,brush)
            self.setColorTransparency(painter, color,-1)
            painter.drawPath(path)

    def drawtrack(self, shape, painter, color):
        if self.canvas.parent.drawObjectTrack():
            track = self.canvas.parent.tracking.getObjectTrack(shape.objectNumber)
            if track and len(track) > 1:
                tp = self.canvas.parent.files.currentFrame if self.canvas.parent.tracking.stackMode else self.canvas.parent.files.currentImage
                track = track[:tp+1]
                path = QPainterPath()  
                fp = next(x[0] for x in enumerate(track) if x[1] != [-1,-1])
                path.moveTo(self.canvas.np2QPoint(track[fp]))
                [path.lineTo(self.canvas.np2QPoint(x)) for x in track[1:] if x != [-1,-1]]
                self.setColorTransparency(painter, color,-1)
                painter.drawPath(path)
                

    def contour2Path(self, contour):
        path = QPainterPath()
        if contour.points is not None:
            path.moveTo(self.canvas.np2QPoint(contour.points[0].reshape(2,1)))
            [path.lineTo(self.canvas.np2QPoint(x.reshape(2,1))) for x in contour.points]
            path.lineTo(self.canvas.np2QPoint(contour.points[0].reshape(2,1)))
            
            innerpath = QPainterPath()
            for c in contour.innercontours:           
                innerpath.moveTo(self.canvas.np2QPoint(c[0].reshape(2,1)))
                [innerpath.lineTo(self.canvas.np2QPoint(p.reshape(2,1))) for p in c]
                innerpath.lineTo(self.canvas.np2QPoint(c[0].reshape(2,1)))

            path = path.subtracted(innerpath)
        return path
    

    def addPoint2NewContour(self, p_image):
        p_last = self.canvas.np2QPoint(self.NewContour.getLastPoint())
        if (p_image != p_last):
            self.NewContour.addPoint(self.canvas.QPoint2np(p_image))
            self.addline(p_last,p_image)
            
  
    def prepareNewContour(self):
        if self.canvas.hasImage():
            width = self.canvas.image().width()
            height = self.canvas.image().height()

            self.contoursketch = np.zeros((height, width), np.uint8) 
            Contour.drawContoursToImage(self.contoursketch, self.contours.getShapesOfClass_x(self.canvas.parent.activeClass()))


    def getFinalContours(self):
        if self.sketch is None:
            return
            
        contours = Contour.extractContoursFromImage(self.contoursketch, not self.canvas.parent.allowInnerContours)
        # delete changed contours
        old_contours = self.contours.getShapesOfClass_x(self.canvas.parent.activeClass())
        changedcontours = Contour.getContoursNotinListOfContours(old_contours,contours)
        self.contours.deleteShapes(changedcontours)
            
        for c in contours:
            c.setClassLabel(self.canvas.parent.activeClass())
            
        self.contours.addShapes(contours)
        self.canvas.redrawImage()

        
    def finishNewContour(self, delete = False, close = True, drawaspolygon = False):
        if self.NewContour is not None:
            if close:
                self.NewContour.closeContour()
            i = 0 if delete else 1
            if self.NewContour.numPoints() > 1:
                if drawaspolygon:
                    cv2.polylines(self.sketch, [self.NewContour.points], 0, (i), thickness=max(self.canvas.pen_size,1), lineType =cv2.LINE_4)  
                else:
                    cv2.drawContours(self.sketch, [self.NewContour.points], 0, (i), -1) 
        self.getFinalContours()
        self.NewContour = None

    def drawcontouraccessories(self, contour, painter = None):
        color = self.canvas.parent.ClassColor(contour.classlabel)
        if painter is None:
            painter = self.getPainter(color = color)
        self.setPainterColor(painter, color)
        if self.canvas.drawShapeNumber:
            contournumber = self.contours.getShapeNumber(contour)
            painter.drawText(self.getLabelNumPosition(contour) , str(contournumber))
            
        if self.canvas.drawSkeleton:
            skeleton = contour.getSkeleton(self.canvas.skeletonsmoothingfactor)         
            
            skelpath = QPainterPath()
            if skeleton is not None:
                for cnt in skeleton:
                    i = 0
                    for c in cnt:         
                        if i == 0:
                            skelpath.moveTo(c[0][0], c[0][1])
                        else:
                            skelpath.lineTo(c[0][0], c[0][1])
                        i = i+1
                col = color.darker(65)
                self.setPainterColor(painter, col)
                self.setColorTransparency(painter, color, transparency = -1)
                painter.drawPath(skelpath)

        
    def getLabelNumPosition(self, contour):
        bp = contour.getBottomPoint()
        pos = QPoint(bp[0],bp[1])+QPoint(0,self.canvas.FontSize + 10)
        if pos.y() > self.canvas.image().height():
            pos.setY(self.canvas.image().height() - self.canvas.FontSize - 5)
        if pos.x() > self.canvas.image().width():
            pos.setX(self.canvas.image().width() - self.canvas.FontSize - 5)
        if pos.x() < 0:
            pos.setX(5)
        return pos


# -*- coding: utf-8 -*-
"""
Created on Mon Jun 29 16:11:55 2020

@author: Koerber
"""


from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from ui.painter.Painter import Painter
import utils.shapes.Point as Point
import os
from ui.Tools import canvasTool
import copy

class ObjectPainter(Painter):
    def __init__(self, canvas):
        super(ObjectPainter,self).__init__(canvas)
        self.points = Point.Points()
        self.backgroundmode = False
        self.tools.append(canvasTool.objectnumber)
        self.tools.append(canvasTool.objectcolor)
        self.tools.append(canvasTool.deleteobject)


    @property
    def shapes(self):
        if self.backgroundmode:
            return self.contours
        else:
            return self.points
        
    def clearFOV(self):
        to_del = []
        fov = self.canvas.getFieldOfViewRect()
        for p in self.points.shapes:
            if fov.x() < p.coordinates[0] < fov.x()+fov.width() and fov.y() < p.coordinates[1] < fov.y()+fov.height():
                to_del.append(p)
        self.points.deleteShapes(to_del)
        self.canvas.redrawImage()
        
    def addShapes(self,shapes):
        self.points.addShapes(shapes)

    def enableDrawBackgroundMode(self):
        super(ObjectPainter, self).enableDrawBackgroundMode()
        self.backgroundmode = True

    def disableDrawBackgroundMode(self):
        super(ObjectPainter, self).disableDrawBackgroundMode()
        self.tools.append(canvasTool.drag)
        self.tools.append(canvasTool.point)
        self.tools.append(canvasTool.shift)
        self.tools.append(canvasTool.assign)
        self.tools.append(canvasTool.assist)
        self.tools.append(canvasTool.delete)
        self.tools.append(canvasTool.objectnumber)
        self.tools.append(canvasTool.objectcolor)
        self.tools.append(canvasTool.deleteobject)
        self.backgroundmode = False
        
    def draw(self):
        super(ObjectPainter, self).draw()
        # for x in range(self.canvas.parent.NumOfClasses()):
        for x in range(self.points.getMaxClass()):    
            self.drawClass(x+1)

    def drawClass(self, x):  
        painter = self.getPainter()
        path = QPainterPath()
        shapes = self.points.getShapesOfClass_x(x)
        color = self.canvas.parent.ClassColor(x)
        self.setPainterColor(painter, color)
        drawall = True

        for p in shapes:
            qp = self.canvas.Point2QPoint(p)
            p1 = QPoint(qp.x() - self.canvas.FontSize, qp.y())
            p2 = QPoint(qp.x() + self.canvas.FontSize, qp.y())
            p3 = QPoint(qp.x(), qp.y() - self.canvas.FontSize)
            p4 = QPoint(qp.x(), qp.y() + self.canvas.FontSize)
                     
            path.moveTo(p1)
            path.lineTo(p2)
            path.moveTo(p3)
            path.lineTo(p4)

            if self.canvas.parent.colorCodeObjects():
                if p.objectNumber != -1:
                    color = self.getObjectColor(p.objectNumber)
                else:
                    color = self.canvas.parent.ClassColor(x)
                self.setPainterColor(painter, color)
                painter.drawPath(path) 
                path = QPainterPath()
                drawall = False

            self.drawtrack(p, painter, color)
            if self.canvas.drawShapeNumber:
                if not drawall:
                    color = self.canvas.parent.ClassColor(x)
                    self.setPainterColor(painter, color)
                pointnumber = self.points.getShapeNumber(p)
                painter.drawText(self.getLabelNumPosition(p) , str(pointnumber))
            
        if drawall:
            painter.drawPath(path)   
    
    def clear(self):
        super(ObjectPainter, self).clear()
        self.points.clear()

    def load(self):
        self.clear()
        path = self.canvas.parent.files.CurrentLabelPath()
        if path is not None and os.path.exists(path):
            points, background = Point.loadPoints(path)
            self.points.addShapes(points)
            if background:
                self.contours.addShapes(background)
        
    def save(self):
        if len(self.trackChanges) > self.maxChanges:
            self.trackChanges.pop(0)
        self.trackChanges.append((copy.deepcopy(self.points.shapes), copy.deepcopy(self.backgroundshapes)))
        self.canvas.parent.checkIfUndoPossible()
        if not self.points.empty() or not self.contours.empty():
            self.canvas.parent.ensureLabelFolder()
            Point.savePoints(self.points.shapes, self.canvas.parent.files.CurrentLabelPath(), background = self.backgroundshapes)
        else:
            self.canvas.parent.deleteLabel()

    def undo(self):
        if len(self.trackChanges) < 2:
            return
        self.clear()
        points, background = self.trackChanges[-2]
        self.trackChanges.pop()
        self.trackChanges.pop()
        self.points.addShapes(points)
        if background:
            self.contours.addShapes(background)
        self.canvas.parent.checkIfUndoPossible()

        
    def checkForChanges(self):
        # to do
        self.canvas.redrawImage()
        
        
    def addPoint(self, point, classlabel=None):
        if classlabel == None:
            classlabel = self.canvas.parent.activeClass()
        
        self.points.addShape(Point.Point(classlabel,self.canvas.QPoint2npPoint(point)))  
        self.canvas.redrawImage()

    def autodetection(self):
        if self.canvas.hasImage():
            self.clearFOV()
            image = self.canvas.getFieldOfViewImage()
            fov = self.canvas.getFieldOfViewRect()
            # not using self.dl.dataloader.readImage() here as a 8-bit rgb image is required and no extra preprocessing
    
            if image is None:
                self.canvas.parent.PopupWarning('No image loaded')
                return
            try:
                prediction = self.canvas.parent.dl.AutoSegment(image)
            except:
                self.canvas.parent.PopupWarning('Auto prediction not possible')
                return
            shapes = Point.extractPointsFromContours(prediction, 5, offset = (int(fov.x()),int(fov.y())))
            for s in shapes:
                s.setClassLabel(self.canvas.parent.activeClass())
            self.points.addShapes(shapes)
            self.checkForChanges()

            
    def getLabelNumPosition(self, point):
        coo = point.coordinates
        pos = QPoint(coo[0],coo[1])+QPoint(10,self.canvas.FontSize+10)
        if pos.y() > self.canvas.image().height():
            pos.setY(self.canvas.image().height() - self.canvas.FontSize - 5)
        if pos.x() > self.canvas.image().width():
            pos.setX(self.canvas.image().width() - self.canvas.FontSize - 5)
        if pos.x() < 0:
            pos.setX(5)
        return pos


# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 15:37:13 2020

@author: Koerber
"""

from abc import ABC, abstractmethod
from enum import Enum


class dlMetric(Enum):
    iou = 1
    dice_coeff = 2
    pixel_accuracy = 3
    mean_squared_err = 4
    root_mean_squared_err = 5
    mean_abs_err = 6
    mean_squared_log_err = 7
    accuracy = 8
    precision = 9
    recall = 10
    f1 = 11


class Metrics(ABC):
    def __init__(self, parent):
        self.supportedmetrics = []
        self.parent = parent
        self._metric = None
        
    @abstractmethod
    def getMetric(self):
        pass
    
    @property
    def metric(self):
        return self._metric
    
    @metric.setter
    def metric(self, metric):
        self._metric = metric


# -*- coding: utf-8 -*-
"""
Created on Fri Nov  8 09:47:31 2019

@author: Koerber
"""



import cv2
import numpy as np
import random
import threading
from tensorflow.keras.preprocessing.image import ImageDataGenerator
import time
from tensorflow.python.keras.utils.data_utils import Sequence
import dl.utils.dl_utils as dl_utils

class TrainingDataGenerator(Sequence):
    def __init__(self, parent, validation = False):
        self.parent = parent
        self.lock = threading.Lock()
        self.inMemory  = parent.TrainInMemory
        self.validation = validation
        if self.inMemory:
            self.images,self.labels = parent.data.loadTrainingDataSet(validation = self.validation)
        else:
            self.images = self.parent.data.getImagePaths(self.validation)
            self.labels = self.parent.data.getLabelPaths(self.validation)
            self.parent.data.getAutoClassWeights()
            
        self.parent.augmentation.initAugmentation()
        self.indices = self.parent.data.getTileIndices(validation, equalTilesperImage = not self.inMemory)
        # self.numImages = len(self.images)
        # experimental: 
        # calculate the number of tiles as the possible crops from all images
        # note that crops may overlap (avoided for validation)
        # length of generator derived from tiles
        self.numTiles = len(self.indices)
        if not validation:
            random.shuffle(self.indices)

    def __len__(self):
        # return int(np.ceil(self.numImages / float(self.batch_size)))
        return self.parent.data.getNumberOfBatches(self.validation)

    def __getitem__(self, idx):
        # important question: do we need a lock here (slows down training time significantly)
        # so far no problems detected
        #with self.lock:

            batch_start = idx * self.parent.batch_size
            batch_end = min(batch_start + self.parent.batch_size, self.numTiles)
            if self.inMemory:
                batch_images,batch_masks = self._loadBatchFromMemory(batch_start, batch_end)
            else:
                batch_images,batch_masks = self._loadBatchFromDisk(batch_start, batch_end)
            

            img, mask = self.parent.augmentation.augment(batch_images,batch_masks, self.validation)

            img = self.parent.data.preprocessImage(img)   
            mask = self.parent.data.preprocessLabel(mask)

            return img, mask
        
    def _loadBatchFromMemory(self, batch_start, batch_end): 
        # are we sure this create a shallow copy? otherwise it would be terrible inefficient memory use
        batch_images = [self.images[self.indices[i]] for i in range(batch_start,batch_end)]
        batch_masks = [self.labels[self.indices[i]] for i in range(batch_start,batch_end)]

        return batch_images, batch_masks
        
    def _loadBatchFromDisk(self, batch_start, batch_end):
        batch_images = []
        batch_masks = []
        for i in range(batch_start, batch_end):
            train_img, train_mask = self.parent.data.createImageLabelPair(self.indices[i],validation=self.validation)
            batch_images.append(train_img)
            batch_masks.append(train_mask)
        return batch_images, batch_masks
        
        
    def on_epoch_end(self):
        if not self.validation:
            random.shuffle(self.indices)
        

class PredictionDataGenerator(Sequence):
    def __init__(self, parent, image):
        self.parent = parent
        self.lock = threading.Lock()
        self.inMemory  = parent.TrainInMemory

        channels = 1 if self.parent.MonoChrome is True else 3
        self.image = image.reshape(image.shape[0], image.shape[1], channels)
        self.t_width = self.parent.augmentation.outputwidth - self.parent.ModelOverlap
        self.t_height = self.parent.augmentation.outputheight - self.parent.ModelOverlap 
        
        
        if self.image.shape[1] < self.parent.augmentation.outputwidth or self.image.shape[0] < self.parent.augmentation.outputheight:
            w = max(0,self.parent.augmentation.outputwidth - self.image.shape[1])
            h = max(0,self.parent.augmentation.outputheight - self.image.shape[0])
            #if self.image.shape
            pad = ((0,h), (0,w), (0, 0))
            self.image = np.pad(self.image, pad, 'reflect')

        pad = ((self.parent.ModelOverlap//2, self.parent.ModelOverlap//2), (self.parent.ModelOverlap//2, self.parent.ModelOverlap//2), (0, 0))       
        self.image = np.pad(self.image, pad, 'reflect')
           
        self.num_in_width = np.ceil(image.shape[1] / self.t_width)
        self.num_in_height = np.ceil(image.shape[0] / self.t_height)
        self.numTiles = int(self.num_in_width * self.num_in_height)

    def __len__(self):
        return int(np.ceil(self.num_in_width * self.num_in_height / (self.parent.batch_size)))

    def __getitem__(self, idx):
        with self.lock:
            batch_start = idx * self.parent.batch_size
            batch_end = min(batch_start + self.parent.batch_size, self.numTiles)


            batch = []
            for currentTile in range(batch_start,batch_end):      
                w = (currentTile % self.num_in_width) * self.t_width
                h = (currentTile // self.num_in_width) * self.t_height

                w_0 = max(w, 0)
                h_0 = max(h, 0)
                
                # split factor need to be reworked and reasonable, get it from network
                h_til = min(h_0+self.t_height+self.parent.ModelOverlap, self.image.shape[0])
                w_til = min(w_0+self.t_width+self.parent.ModelOverlap,  self.image.shape[1])
                
                # a full tile at the borders results in better segmentations
                if h_til == self.image.shape[0]:
                    h_0 = max(self.image.shape[0] - (self.t_height + self.parent.ModelOverlap), 0)
                if w_til == self.image.shape[1]:
                    w_0 = max(self.image.shape[1] - (self.t_width + self.parent.ModelOverlap), 0)

                tile = self.image[int(h_0):int(h_til), int(w_0):int(w_til)]
                batch.append(tile)
                
            img = self.parent.data.preprocessImage(np.asarray(batch))
            return self.parent.augmentation.checkModelSizeCompatibility(img)



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 19 13:48:32 2020

@author: Koerber
"""

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from ui.style import styleForm
from ui.ui_utils import DCDButton
from dl.method.mode import dlMode


class Window(object):
    def setupUi(self, Form):
        Form.setWindowTitle('Post Processing') 
        styleForm(Form)

        self.centralWidget = QWidget(Form)
        
        self.vlayout = QVBoxLayout(self.centralWidget)
        self.vlayout.setContentsMargins(3, 10, 3, 3)
        self.vlayout.setSpacing(2)
        self.centralWidget.setLayout(self.vlayout)

        self.CBTrackingMode = QCheckBox("Tracking Mode",self.centralWidget)
        self.CBTrackingMode.setToolTip('Check to enable tracking mode')
        self.CBTrackingMode.setObjectName('TrackingMode_od')

        self.BCopyStackLabels = DCDButton(self.centralWidget, "Use as Stack Label")
        self.BCopyStackLabels.setToolTip('Press to reuse current label for all frames in the current stack')

        self.vlayout.addWidget(self.CBTrackingMode)
        self.vlayout.addWidget(self.BCopyStackLabels)

        self.centralWidget.setFixedSize(self.vlayout.sizeHint())
        Form.setFixedSize(self.vlayout.sizeHint())


class ObjectDetectionPostProcessingWindow(QMainWindow, Window):
    def __init__(self, parent ):
        super(ObjectDetectionPostProcessingWindow, self).__init__(parent)
        self.parent = parent
        self.setupUi(self)
        self.CBTrackingMode.stateChanged.connect(self.enableTrackingMode)
        self.CBTrackingMode.setChecked(self.parent.TrackingModeEnabled)
        self.BCopyStackLabels.clicked.connect(self.shareStackLabels)

    def enableTrackingMode(self):
        if self.parent.LearningMode() == dlMode.Object_Counting:
            self.parent.enableTrackingMode(self.CBTrackingMode.isChecked())

    def shareStackLabels(self):
        self.parent.copyStackLabels()





# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 16:43:11 2020

@author: Koerber
"""

from dl.metric.metrics import Metrics, dlMetric
from dl.metric.metric_functions import *

class ClassificationMetrics(Metrics):
    def __init__(self,parent):
        super(ClassificationMetrics,self).__init__(parent)
        self.supportedmetrics.append(dlMetric.accuracy)
        self.supportedmetrics.append(dlMetric.precision)
        self.supportedmetrics.append(dlMetric.recall)
        self.supportedmetrics.append(dlMetric.f1)
        self.metric = dlMetric.accuracy
        
    def getMetric(self):
        if self.metric == dlMetric.accuracy:
            return self.accuracy_function(binary = self.parent.NumClasses <= 2)
        if self.metric == dlMetric.precision:
            return [precision]
        if self.metric == dlMetric.recall:
            return [recall]
        if self.metric == dlMetric.f1:
            return [f1]
        else:
            raise # not supported


    def accuracy_function(self, binary):
        # return 'accuracy' 
        if binary:
            return [accuracy_binary()]
        else:
            return [accuracy()]


from dl.models.keras_applications import get_submodules_from_kwargs

from ._common_blocks import Conv2dBn
from ._utils import freeze_model
from ..backbones.backbones_factory import Backbones

backend = None
layers = None
models = None
keras_utils = None


# ---------------------------------------------------------------------
#  Utility functions
# ---------------------------------------------------------------------

def get_submodules():
    return {
        'backend': backend,
        'models': models,
        'layers': layers,
        'utils': keras_utils,
    }


# ---------------------------------------------------------------------
#  Blocks
# ---------------------------------------------------------------------

def Conv3x3BnReLU(filters, use_batchnorm, name=None):
    kwargs = get_submodules()

    def wrapper(input_tensor):
        return Conv2dBn(
            filters,
            kernel_size=3,
            activation='relu',
            kernel_initializer='he_uniform',
            padding='same',
            use_batchnorm=use_batchnorm,
            name=name,
            **kwargs
        )(input_tensor)

    return wrapper


def DecoderUpsamplingX2Block(filters, stage, use_batchnorm=False):
    up_name = 'decoder_stage{}_upsampling'.format(stage)
    conv1_name = 'decoder_stage{}a'.format(stage)
    conv2_name = 'decoder_stage{}b'.format(stage)
    concat_name = 'decoder_stage{}_concat'.format(stage)

    concat_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    def wrapper(input_tensor, skip=None):
        x = layers.UpSampling2D(size=2, name=up_name)(input_tensor)

        if skip is not None:
            x = layers.Concatenate(axis=concat_axis, name=concat_name)([x, skip])

        x = Conv3x3BnReLU(filters, use_batchnorm, name=conv1_name)(x)
        x = Conv3x3BnReLU(filters, use_batchnorm, name=conv2_name)(x)

        return x

    return wrapper


def DecoderTransposeX2Block(filters, stage, use_batchnorm=False):
    transp_name = 'decoder_stage{}a_transpose'.format(stage)
    bn_name = 'decoder_stage{}a_bn'.format(stage)
    relu_name = 'decoder_stage{}a_relu'.format(stage)
    conv_block_name = 'decoder_stage{}b'.format(stage)
    concat_name = 'decoder_stage{}_concat'.format(stage)

    concat_axis = bn_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    def layer(input_tensor, skip=None):

        x = layers.Conv2DTranspose(
            filters,
            kernel_size=(4, 4),
            strides=(2, 2),
            padding='same',
            name=transp_name,
            use_bias=not use_batchnorm,
        )(input_tensor)

        if use_batchnorm:
            x = layers.BatchNormalization(axis=bn_axis, name=bn_name)(x)

        x = layers.Activation('relu', name=relu_name)(x)

        if skip is not None:
            x = layers.Concatenate(axis=concat_axis, name=concat_name)([x, skip])

        x = Conv3x3BnReLU(filters, use_batchnorm, name=conv_block_name)(x)

        return x

    return layer


# ---------------------------------------------------------------------
#  Unet Decoder
# ---------------------------------------------------------------------

def build_unet(
        backbone,
        decoder_block,
        skip_connection_layers,
        decoder_filters=(256, 128, 64, 32, 16),
        n_upsample_blocks=5,
        classes=1,
        activation='sigmoid',
        use_batchnorm=True,
):
    input_ = backbone.input
    x = backbone.output

    # extract skip connections
    skips = ([backbone.get_layer(name=i).output if isinstance(i, str)
              else backbone.get_layer(index=i).output for i in skip_connection_layers])

    # add center block if previous operation was maxpooling (for vgg models)
    if isinstance(backbone.layers[-1], layers.MaxPooling2D):
        x = Conv3x3BnReLU(512, use_batchnorm, name='center_block1')(x)
        x = Conv3x3BnReLU(512, use_batchnorm, name='center_block2')(x)

    # building decoder blocks
    for i in range(n_upsample_blocks):

        if i < len(skips):
            skip = skips[i]
        else:
            skip = None

        x = decoder_block(decoder_filters[i], stage=i, use_batchnorm=use_batchnorm)(x, skip)

    # model head (define number of output classes)
    x = layers.Conv2D(
        filters=classes,
        kernel_size=(3, 3),
        padding='same',
        use_bias=True,
        kernel_initializer='glorot_uniform',
        name='final_conv',
    )(x)
    x = layers.Activation(activation, name=activation)(x)

    # create keras model instance
    model = models.Model(input_, x)

    return model


# ---------------------------------------------------------------------
#  Unet Model
# ---------------------------------------------------------------------

def Unet(
        backbone_name='vgg16',
        input_shape=(None, None, 3),
        classes=1,
        activation='sigmoid',
        weights=None,
        encoder_weights='imagenet',
        encoder_freeze=False,
        encoder_features='default',
        decoder_block_type='upsampling',
        decoder_filters=(256, 128, 64, 32, 16),
        decoder_use_batchnorm=True,
        **kwargs
):
    """ Unet_ is a fully convolution neural network for image semantic segmentation

    Args:
        backbone_name: name of classification model (without last dense layers) used as feature
            extractor to build segmentation model.
        input_shape: shape of input data/image ``(H, W, C)``, in general
            case you do not need to set ``H`` and ``W`` shapes, just pass ``(None, None, C)`` to make your model be
            able to process images af any size, but ``H`` and ``W`` of input images should be divisible by factor ``32``.
        classes: a number of classes for output (output shape - ``(h, w, classes)``).
        activation: name of one of ``keras.activations`` for last model layer
            (e.g. ``sigmoid``, ``softmax``, ``linear``).
        weights: optional, path to model weights.
        encoder_weights: one of ``None`` (random initialization), ``imagenet`` (pre-training on ImageNet).
        encoder_freeze: if ``True`` set all layers of encoder (backbone model) as non-trainable.
        encoder_features: a list of layer numbers or names starting from top of the model.
            Each of these layers will be concatenated with corresponding decoder block. If ``default`` is used
            layer names are taken from ``DEFAULT_SKIP_CONNECTIONS``.
        decoder_block_type: one of blocks with following layers structure:

            - `upsampling`:  ``UpSampling2D`` -> ``Conv2D`` -> ``Conv2D``
            - `transpose`:   ``Transpose2D`` -> ``Conv2D``

        decoder_filters: list of numbers of ``Conv2D`` layer filters in decoder blocks
        decoder_use_batchnorm: if ``True``, ``BatchNormalisation`` layer between ``Conv2D`` and ``Activation`` layers
            is used.

    Returns:
        ``keras.models.Model``: **Unet**

    .. _Unet:
        https://arxiv.org/pdf/1505.04597

    """

    global backend, layers, models, keras_utils
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)

    if decoder_block_type == 'upsampling':
        decoder_block = DecoderUpsamplingX2Block
    elif decoder_block_type == 'transpose':
        decoder_block = DecoderTransposeX2Block
    else:
        raise ValueError('Decoder block type should be in ("upsampling", "transpose"). '
                         'Got: {}'.format(decoder_block_type))

    backbone = Backbones.get_backbone(
        backbone_name,
        input_shape=input_shape,
        weights=encoder_weights,
        include_top=False,
        **kwargs,
    )

    if encoder_features == 'default':
        encoder_features = Backbones.get_feature_layers(backbone_name, n=4)

    model = build_unet(
        backbone=backbone,
        decoder_block=decoder_block,
        skip_connection_layers=encoder_features,
        decoder_filters=decoder_filters,
        classes=classes,
        activation=activation,
        n_upsample_blocks=len(decoder_filters),
        use_batchnorm=decoder_use_batchnorm,
    )

    # lock encoder weights for fine-tuning
    if encoder_freeze:
        freeze_model(backbone, **kwargs)

    # loading model weights
    if weights is not None:
        model.load_weights(weights)

    return model



from math import ceil

from tensorflow.keras.layers import Concatenate, Add
from tensorflow.keras.layers import AveragePooling2D, ZeroPadding2D
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Layer


import tensorflow.keras.backend as K
import tensorflow as tf


class Upsampling(Layer):

    def __init__(self, new_size, **kwargs):
        self.new_size = new_size
        super(Upsampling, self).__init__(**kwargs)

    def build(self, input_shape):
        super(Upsampling, self).build(input_shape)

    def call(self, inputs, **kwargs):
        new_height, new_width = self.new_size
        resized = tf.image.resize(inputs, [new_height, new_width])
        return resized

    def compute_output_shape(self, input_shape):
        return tuple([None, self.new_size[0], self.new_size[1], input_shape[3]])

    def get_config(self):
        config = super(Upsampling, self).get_config()
        config['new_size'] = self.new_size
        return config


def psp_block(prev_layer, level, feature_map_shape, input_shape):
    if input_shape == (512, 512):
        kernel_strides_map = {1: [64, 64],
                              2: [32, 32],
                              3: [22, 21],
                              6: [11, 9]}  # TODO: Level 6: Kernel correct, but stride not exactly the same as Pytorch
    else:
        raise ValueError("Pooling parameters for input shape " + input_shape + " are not defined.")

    if K.image_data_format() == 'channels_last':
        bn_axis = 3
    else:
        bn_axis = 1

    names = [
        "class_psp_" + str(level) + "_conv",
        "class_psp_" + str(level) + "_bn"
    ]
    kernel = (kernel_strides_map[level][0], kernel_strides_map[level][0])
    strides = (kernel_strides_map[level][1], kernel_strides_map[level][1])
    prev_layer = AveragePooling2D(kernel, strides=strides)(prev_layer)
    prev_layer = Conv2D(512, (1, 1), strides=(1, 1), name=names[0], use_bias=False)(prev_layer)
    prev_layer = BN(bn_axis, name=names[1])(prev_layer)
    prev_layer = Activation('relu')(prev_layer)
    prev_layer = Upsampling(feature_map_shape)(prev_layer)
    return prev_layer


def build_pyramid_pooling_module(res, input_shape, nb_classes, sigmoid=False, output_size=None):
    """Build the Pyramid Pooling Module."""
    # ---PSPNet concat layers with Interpolation
    feature_map_size = tuple(int(ceil(input_dim / 8.0)) for input_dim in input_shape)
    if K.image_data_format() == 'channels_last':
        bn_axis = 3
    else:
        bn_axis = 1

    interp_block1 = psp_block(res, 1, feature_map_size, input_shape)
    interp_block2 = psp_block(res, 2, feature_map_size, input_shape)
    interp_block3 = psp_block(res, 3, feature_map_size, input_shape)
    interp_block6 = psp_block(res, 6, feature_map_size, input_shape)

    # concat all these layers. resulted
    res = Concatenate()([interp_block1,
                         interp_block2,
                         interp_block3,
                         interp_block6,
                         res])
    x = Conv2D(512, (1, 1), strides=(1, 1), padding="same", name="class_psp_reduce_conv", use_bias=False)(res)
    x = BN(bn_axis, name="class_psp_reduce_bn")(x)
    x = Activation('relu')(x)

    x = Conv2D(nb_classes, (1, 1), strides=(1, 1), name="class_psp_final_conv")(x)

    if output_size:
        x = Upsampling(output_size)(x)

    if sigmoid:
        x = Activation('sigmoid')(x)
    return x


def BN(axis, name=""):
    return BatchNormalization(axis=axis, momentum=0.1, name=name, epsilon=1e-5)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 28 16:36:28 2020

@author: Koerber
"""


import cv2
import numpy as np

class GrabCutSegmentation():
    def __init__(self):
        self.bg_Model = np.zeros((1,65),np.float64)
        self.fg_Model = np.zeros((1,65),np.float64)
        self.iterations = 1

    
    def trainAndSegment(self,image, mask):
        if np.any(mask == 1):
            # no samples       
            mask, self.bg_Model, self.fg_Model = cv2.grabCut(image,mask,None,self.bg_Model,self.fg_Model,self.iterations,cv2.GC_INIT_WITH_MASK)
        retmask = np.where((mask==2)|(mask==0),0,1).astype('uint8')
        return retmask

    def Segment(self,image):
        mask = np.zeros(image.shape[:2],np.uint8)
        mask[mask==0] = 2
        mask, self.bg_Model, self.fg_Model = cv2.grabCut(image,mask,None,self.bg_Model,self.fg_Model,1,cv2.GC_EVAL_FREEZE_MODEL)
        retmask = np.where((mask==2)|(mask==0),0,1).astype('uint8')
        return retmask


# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 14:30:54 2020

@author: Koerber
"""

from dl.loss.losses import Losses, dlLoss
from dl.loss.loss_functions import *

class RegressionLosses(Losses):
    def __init__(self,parent):
        super(RegressionLosses,self).__init__(parent)
        self.supportedlosses.append(dlLoss.mean_squared_err)
        self.supportedlosses.append(dlLoss.mean_squared_log_err)
        self.supportedlosses.append(dlLoss.mean_abs_err)
        self.loss = dlLoss.mean_squared_err

    def getLoss(self):
        if self.loss == dlLoss.mean_squared_err:
            return mean_squared_error(class_weights = self.parent.data.class_weights)
        if self.loss == dlLoss.mean_abs_err:
            return mean_absolute_error(class_weights = self.parent.data.class_weights)
        if self.loss == dlLoss.mean_squared_log_err:
            return mean_squared_logarithmic_error(class_weights = self.parent.data.class_weights)
        else:
            raise # not supported
            

        
        